from typing import Callable, Sequence, Type
import numpy as np
import astropy.units as u
import named_arrays as na
import named_arrays._scalars.scalar_array_functions
from . import vectors

__all__ = [
    'SINGLE_ARG_FUNCTIONS',
    'ARRAY_CREATION_LIKE_FUNCTIONS',
    'SEQUENCE_FUNCTIONS',
    'DEFAULT_FUNCTIONS',
    'PERCENTILE_LIKE_FUNCTIONS',
    'ARG_REDUCE_FUNCTIONS',
    'FFT_LIKE_FUNCTIONS',
    'FFTN_LIKE_FUNCTIONS',
    'HANDLED_FUNCTIONS',
    'STACK_LIKE_FUNCTIONS',
]

SINGLE_ARG_FUNCTIONS = named_arrays._scalars.scalar_array_functions.SINGLE_ARG_FUNCTIONS
ARRAY_CREATION_LIKE_FUNCTIONS = named_arrays._scalars.scalar_array_functions.ARRAY_CREATION_LIKE_FUNCTIONS
SEQUENCE_FUNCTIONS = named_arrays._scalars.scalar_array_functions.SEQUENCE_FUNCTIONS
DEFAULT_FUNCTIONS = [
    np.ndim,
    np.min,
    np.nanmin,
    np.max,
    np.nanmax,
    np.sum,
    np.nansum,
    np.prod,
    np.nanprod,
    np.mean,
    np.nanmean,
    np.std,
    np.nanstd,
    np.var,
    np.nanvar,
    np.median,
    np.nanmedian,
    np.all,
    np.any,
    np.ptp,
    np.count_nonzero,
]
PERCENTILE_LIKE_FUNCTIONS = [
    np.percentile,
    np.nanpercentile,
    np.quantile,
    np.nanquantile,
]
ARG_REDUCE_FUNCTIONS = [
    np.argmin,
    np.nanargmin,
    np.argmax,
    np.nanargmax,
]
FFT_LIKE_FUNCTIONS = [
    np.fft.fft,
    np.fft.ifft,
    np.fft.rfft,
    np.fft.irfft,
]
FFTN_LIKE_FUNCTIONS = [
    np.fft.fft2,
    np.fft.ifft2,
    np.fft.rfft2,
    np.fft.irfft2,
    np.fft.fftn,
    np.fft.ifftn,
    np.fft.rfftn,
    np.fft.irfftn,
]
EMATH_FUNCTIONS = named_arrays._scalars.scalar_array_functions.EMATH_FUNCTIONS
STACK_LIKE_FUNCTIONS = [
    np.stack,
    np.concatenate,
]
HANDLED_FUNCTIONS = dict()


def array_functions_single_arg(
    func: Callable,
    a: na.AbstractVectorArray,
) -> na.AbstractExplicitVectorArray:
    components = a.components
    components = {c: func(components[c]) for c in components}
    return a.type_explicit.from_components(components)


def array_function_array_creation_like(
        func: Callable,
        prototype: na.AbstractVectorArray,
        dtype: None | type | np.dtype = None,
        order: str = "K",
        subok: bool = True,
        shape: dict[str, int] = None,
) -> na.AbstractVectorArray:

    components = prototype.components

    components_result = dict()
    for c in components:
        components_result[c] = func(
            na.as_named_array(components[c]),
            dtype=dtype,
            order=order,
            subok=subok,
            shape=shape,
        )

    return prototype.type_explicit.from_components(components_result)


def array_function_sequence(
        func: Callable,
        *args: float | u.Quantity | na.AbstractScalar | na.AbstractVectorArray,
        axis: str | na.AbstractVectorArray,
        num: int | na.AbstractVectorArray = 50,
        endpoint: bool = True,
        dtype: None | type | np.dtype = None,
        **kwargs: float | u.Quantity | na.AbstractScalar | na.AbstractVectorArray,
) -> na.AbstractVectorArray:

    try:
        prototype = vectors._prototype(*args, *kwargs.values())
        args = tuple(vectors._normalize(arg, prototype) for arg in args)
        axis = vectors._normalize(axis, prototype)
        num = vectors._normalize(num, prototype)
        kwargs = {k: vectors._normalize(kwargs[k], prototype) for k in kwargs}
    except na.VectorTypeError:
        return NotImplemented

    components_args = tuple(arg.components for arg in args)
    components_axis = axis.components
    components_num = num.components
    components_kwargs = {k: kwargs[k].components for k in kwargs}

    components = {
        c: func(
            *tuple(na.as_named_array(components_arg[c]) for components_arg in components_args),
            axis=components_axis[c],
            num=components_num[c],
            endpoint=endpoint,
            dtype=dtype,
            **{k: na.as_named_array(components_kwargs[k][c]) for k in components_kwargs},
        )
        for c in prototype.components
    }

    return prototype.type_explicit.from_components(components)


def array_function_default(
        func: Callable,
        a: na.AbstractVectorArray,
        axis: None | str | Sequence[str] = None,
        dtype: None | type | np.dtype = np._NoValue,
        out: None | na.AbstractExplicitVectorArray = None,
        keepdims: bool = False,
        initial: None | bool | int | float | complex | u.Quantity = np._NoValue,
        where: na.AbstractScalarArray | na.AbstractVectorArray = np._NoValue,
) -> na.AbstractExplicitVectorArray:

    a = a.explicit
    shape = na.shape_broadcasted(a, where)

    axis_normalized = tuple(shape) if axis is None else (axis, ) if isinstance(axis, str) else axis

    if axis is not None:
        if not set(axis_normalized).issubset(shape):
            raise ValueError(
                f"the `axis` argument, {axis}, must be `None` or a subset of the broadcasted shape of `a` and "
                f"`where`, {shape} "
            )

    shape_base = {ax: shape[ax] for ax in axis_normalized}

    components = a.components
    components_out = out.components if isinstance(out, na.AbstractVectorArray) else {c: out for c in components}
    components_where = where.components if isinstance(where, na.AbstractVectorArray) else {c: where for c in components}

    kwargs_base = dict(
        axis=axis,
        keepdims=keepdims,
    )

    if dtype is not np._NoValue:
        kwargs_base["dtype"] = dtype
    if initial is not np._NoValue:
        kwargs_base["initial"] = initial

    result = a.prototype_vector
    for c in components:
        component = na.as_named_array(components[c])
        where_c = components_where[c]
        shape_c = na.broadcast_shapes(component.shape, na.shape(where_c), shape_base)
        kwargs = dict()
        if where_c is not np._NoValue:
            kwargs["where"] = where_c.broadcast_to(shape_c) if isinstance(where_c, na.AbstractArray) else where_c
        result.components[c] = func(
            component.broadcast_to(shape_c),
            out=components_out[c],
            **kwargs_base,
            **kwargs,
        )

    if out is not None:
        result = out

    return result


def array_function_percentile_like(
        func: Callable,
        a: na.AbstractVectorArray,
        q: float | u.Quantity | na.AbstractScalarArray | na.AbstractVectorArray,
        axis: None | str | Sequence[str] = None,
        out: None | na.AbstractExplicitVectorArray = None,
        overwrite_input: bool = False,
        method: str = 'linear',
        keepdims: bool = False,
) -> na.AbstractExplicitVectorArray:

    a = a.explicit
    shape = a.shape

    axis_normalized = na.axis_normalized(a, axis)

    if axis is not None:
        if not set(axis_normalized).issubset(shape):
            raise ValueError(
                f"the `axis` argument, {axis}, must be `None` or a subset of the shape of `a`, {shape}"
            )

    shape_base = {ax: shape[ax] for ax in axis_normalized}

    components = a.components
    components_q = q.components if isinstance(q, na.AbstractVectorArray) else {c: q for c in components}
    components_out = out.components if isinstance(out, na.AbstractVectorArray) else {c: out for c in components}

    kwargs_base = dict(
        axis=axis,
        overwrite_input=overwrite_input,
        method=method,
        keepdims=keepdims,
    )

    result = a.prototype_vector
    for c in components:
        component = na.as_named_array(components[c])
        shape_c = na.broadcast_shapes(component.shape, shape_base)
        result.components[c] = func(
            component.broadcast_to(shape_c),
            q=components_q[c],
            out=components_out[c],
            **kwargs_base,
        )

    if out is not None:
        result = out

    return result


def array_function_arg_reduce(
        func: Callable,
        a: na.AbstractVectorArray,
        axis: None | str | Sequence[str] = None,
) -> dict[str, na.AbstractVectorArray]:

    a = a.explicit
    components = a.components

    if axis is not None:
        if isinstance(axis, str):
            axis = (axis, )
        if not set(axis).issubset(a.axes):
            raise ValueError(f"Reduction axes {axis} are not a subset of the array axes {a.axes}")
    else:
        if not a.shape:
            raise ValueError(f"Applying {func} to zero-dimensional arrays is not supported")

    axis_normalized = na.axis_normalized(a, axis)

    components_result = dict()
    for c in components:
        component = na.as_named_array(components[c])
        shape_c = component.shape
        axis_c = tuple(ax for ax in axis_normalized if ax in shape_c)
        if axis_c:
            components_result[c] = func(
                a=component,
                axis=axis_c,
            )
        else:
            components_result[c] = dict()

    axes_result = tuple(set(ax for c in components_result for ax in components_result[c]))

    result = dict()
    for ax in axes_result:
        result[ax] = a.prototype_vector
        for c in components:
            result[ax].components[c] = components_result[c].get(ax, None)

    return result


def array_function_fft_like(
        func: Callable,
        a: na.AbstractVectorArray,
        axis: tuple[str, str],
        n: None | int = None,
        norm: str = "backward"
) -> na.AbstractExplicitVectorArray:

    a = a.explicit
    components = a.components
    shape = a.shape

    if axis[0] not in shape:
        raise ValueError(f"`axis` {axis[0]} not in array with shape {shape}")

    shape_axis = {axis[0]: shape[axis[0]]}

    components = {
        c: na.broadcast_to(components[c], na.broadcast_shapes(na.shape(components[c]), shape_axis))
        for c in components
    }

    result = a.prototype_vector
    for c in components:
        result.components[c] = func(
            a=components[c],
            axis=axis,
            n=n,
            norm=norm,
        )

    return result


def array_function_fftn_like(
        func: Callable,
        a: na.AbstractVectorArray,
        axes: dict[str, str],
        s: None | dict[str, int] = None,
        norm: str = "backward",
) -> na.AbstractExplicitVectorArray:

    a = a.explicit
    components = a.components
    shape_a = a.shape

    if not set(axes).issubset(shape_a):
        raise ValueError(f"`axes`, {tuple(axes)}, not a subset of array axes, {tuple(shape_a)}")

    shape_base = {ax: shape_a[ax] for ax in axes}

    result = a.prototype_vector
    for c in components:
        component = na.as_named_array(components[c])
        result.components[c] = func(
            a=na.broadcast_to(component, na.broadcast_shapes(component.shape, shape_base)),
            axes=axes,
            s=s,
            norm=norm,
        )

    return result


def array_function_emath(
    func: Callable,
    *args: na.AbstractScalar | na.AbstractVectorArray,
    **kwargs: na.AbstractScalar | na.AbstractVectorArray,
):
    try:
        prototype = vectors._prototype(*args, *kwargs.values())
        args = tuple(vectors._normalize(a, prototype) for a in args)
        kwargs = {k: vectors._normalize(kwargs[k], prototype) for k in kwargs}
    except vectors.VectorTypeError:     # pragma: nocover
        return NotImplemented

    args = tuple(a.components for a in args)
    kwargs = {k: kwargs[k].components for k in kwargs}

    result = {
        c: func(*[a[c] for a in args], **{k: kwargs[k][c] for k in kwargs})
        for c in prototype.components
    }

    return prototype.type_explicit.from_components(result)


def implements(numpy_function: Callable):
    """Register an ``__array_function__`` implementation for :class:`AbstractVectorArray` objects."""
    def decorator(func):
        HANDLED_FUNCTIONS[numpy_function] = func
        return func
    return decorator


@implements(np.copyto)
def copyto(
        dst: na.AbstractExplicitVectorArray,
        src: na.AbstractScalar | na.AbstractVectorArray,
        casting: str = "same_kind",
        where: bool | na.AbstractScalar | na.AbstractVectorArray = True,
) -> None:
    if not isinstance(dst, na.AbstractExplicitVectorArray):
        return NotImplemented

    components_dst = dst.components

    if isinstance(src, na.AbstractArray):
        if isinstance(src, na.AbstractVectorArray):
            if src.type_abstract == dst.type_abstract:
                components_src = src.components
            else:
                return NotImplemented
        elif isinstance(src, na.AbstractScalar):
            components_src = {c: src for c in components_dst}
        else:
            return NotImplemented
    else:
        components_src = {c: src for c in components_dst}

    if isinstance(where, na.AbstractArray):
        if isinstance(where, na.AbstractVectorArray):
            if where.type_abstract == where.type_explicit:
                components_where = where.components
            else:
                return NotImplemented
        elif isinstance(where, na.AbstractScalar):
            components_where = {c: where for c in components_dst}
        else:
            return NotImplemented
    else:
        components_where = {c: where for c in components_dst}

    for c in components_dst:
        try:
            np.copyto(
                dst=components_dst[c],
                src=components_src[c],
                casting=casting,
                where=components_where[c],
            )
        except TypeError:
            components_dst[c] = components_src[c]


@implements(np.broadcast_to)
def broadcast_to(
        array: na.AbstractVectorArray,
        shape: dict[str, int],
) -> na.AbstractExplicitVectorArray:
    components = array.components
    components_result = {c: na.broadcast_to(array=components[c], shape=shape) for c in components}
    return array.type_explicit.from_components(components_result)


@implements(np.transpose)
def transpose(
        a: na.AbstractVectorArray,
        axes: None | Sequence[str] = None,
) -> na.AbstractExplicitVectorArray:

    components = a.broadcasted.components
    return a.type_explicit.from_components({c: np.transpose(a=components[c], axes=axes) for c in components})


@implements(np.moveaxis)
def moveaxis(
        a: na.AbstractVectorArray,
        source: str | Sequence[str],
        destination: str | Sequence[str],
) -> na.AbstractExplicitVectorArray:

    axes = a.axes

    if isinstance(source, str):
        source = (source, )
    if isinstance(destination, str):
        destination = (destination, )

    set_axis_diff = set(source) - set(axes)
    if set_axis_diff:
        raise ValueError(f"source axes {tuple(set_axis_diff)} not in array axes {axes}")

    components = a.components
    components_result = dict()
    for c in components:
        shape_c = na.shape(components[c])
        source_and_destination = tuple((src, dest) for src, dest in zip(source, destination) if src in shape_c)
        components_result[c] = np.moveaxis(
            a=components[c],
            source=tuple(src_and_dest[0] for src_and_dest in source_and_destination),
            destination=tuple(src_and_dest[1] for src_and_dest in source_and_destination),
        )
    return a.type_explicit.from_components(components_result)


@implements(np.reshape)
def reshape(a: na.AbstractVectorArray, newshape: dict[str, int]) -> na.AbstractExplicitVectorArray:
    components = a.broadcasted.components
    return a.type_explicit.from_components({c: np.reshape(a=components[c], newshape=newshape) for c in components})


def array_function_stack_like(
        func: Callable,
        arrays: Sequence[bool | int | float | complex | str | u.Quantity | na.AbstractScalar | na.AbstractVectorArray],
        axis: str,
        out: None | na.AbstractExplicitVectorArray = None,
        *,
        dtype: str | np.dtype | Type = None,
        casting: str = "same_kind",
) -> na.AbstractExplicitVectorArray:

    for array in arrays:
        if isinstance(array, na.AbstractVectorArray):
            vector_prototype = array
            break

    arrays = [
        vector_prototype.type_explicit.from_scalar(
            scalar=a,
            like=vector_prototype,
        ) if not isinstance(a, na.AbstractVectorArray) else a
        for a in arrays
    ]
    arrays = [a.broadcasted for a in arrays]

    components_arrays = [a.components for a in arrays]

    if out is None:
        components_out = {c: None for c in components_arrays[0]}
    else:
        components_out = out.components

    components_result = dict()
    for c in components_arrays[0]:
        components_result[c] = func(
            [components[c] for components in components_arrays],
            axis=axis,
            out=components_out[c],
            dtype=dtype,
            casting=casting,
        )

    if out is None:
        return vector_prototype.type_explicit.from_components(components_result)
    else:
        return out


@implements(np.sort)
def sort(
        a: na.AbstractVectorArray,
        axis: None | str | Sequence[str],
        kind: None | str = None,
        order: None | str | list[str] = None,
) -> na.AbstractExplicitVectorArray:

    a = a.explicit
    shape_a = a.shape
    components = a.components

    if axis is None:
        axis = tuple(shape_a)
        if not axis:
            return a
    elif isinstance(axis, str):
        axis = (axis, )
    else:
        if not axis:
            raise ValueError(f"if `axis` is a sequence, it must not be empty, got {axis}")

    if not set(axis).issubset(shape_a):
        raise ValueError(f"`axis`, {axis} is not a subset of `a.axes`, {a.axes}")

    shape_base = {ax: shape_a[ax] for ax in axis}

    result = a.prototype_vector
    for c in components:
        component = na.as_named_array(components[c])
        if any(ax in axis for ax in component.axes):
            component = component.broadcast_to(na.broadcast_shapes(component.shape, shape_base))
            result.components[c] = np.sort(component, axis=axis, kind=kind, order=order)
        else:
            result.components[c] = components[c]

    return result


@implements(np.argsort)
def argsort(
        a: na.AbstractVectorArray,
        axis: None | str | Sequence[str],
        kind: None | str = None,
        order: None | str | list[str] = None,
) -> dict[str, na.AbstractExplicitVectorArray]:

    a = a.explicit
    shape_a = a.shape
    components = a.components

    if axis is None:
        axis = tuple(shape_a)
        if not axis:
            return dict()
    elif isinstance(axis, str):
        axis = (axis, )
    else:
        if not axis:
            raise ValueError(f"if `axis` is a sequence, it must not be empty, got {axis}")

    if not set(axis).issubset(shape_a):
        raise ValueError(f"`axis`, {axis} is not a subset of `a.axes`, {a.axes}")

    shape_base = {ax: shape_a[ax] for ax in axis}

    result = {ax: a.prototype_vector for ax in shape_a}
    for c in components:
        component = na.as_named_array(components[c])
        if any(ax in axis for ax in component.axes):
            component = component.broadcast_to(na.broadcast_shapes(component.shape, shape_base))
            result_c = np.argsort(component, axis=axis, kind=kind, order=order)
        else:
            result_c = dict()
        for ax in shape_a:
            result[ax].components[c] = result_c[ax] if ax in result_c else None

    return result


@implements(np.unravel_index)
def unravel_index(
        indices: na.AbstractVectorArray,
        shape: dict[str, int],
) -> dict[str, na.AbstractExplicitVectorArray]:

    indices = indices.explicit
    components = indices.components
    result = {ax: dict() for ax in indices.axes}

    for c in components:
        result_c = np.unravel_index(
            indices=components[c],
            shape=shape,
        )
        for ax in result_c:
            result[ax][c] = result_c[ax]

    result = {ax: indices.type_explicit.from_components(result[ax]) for ax in result}
    return result


@implements(np.array_equal)
def array_equal(
        a1: na.AbstractVectorArray,
        a2: na.AbstractVectorArray,
        equal_nan: bool = False,
) -> bool:

    if not a1.type_abstract == a2.type_abstract:
        return False

    components_a1 = a1.components
    components_a2 = a2.components

    result = True
    for c in components_a1:
        result = result and np.array_equal(
            a1=components_a1[c],
            a2=components_a2[c],
            equal_nan=equal_nan,
        )

    return result


@implements(np.array_equiv)
def array_equiv(
        a1: na.AbstractVectorArray,
        a2: na.AbstractVectorArray,
) -> bool:

    if not a1.type_abstract == a2.type_abstract:
        return False

    components_a1 = a1.components
    components_a2 = a2.components

    result = True
    for c in components_a1:
        result = result and np.array_equiv(
            a1=components_a1[c],
            a2=components_a2[c],
        )

    return result


@implements(np.allclose)
def allclose(
        a: na.ArrayLike,
        b: na.ArrayLike,
        rtol: float = 1e-05,
        atol: float = 1e-08,
        equal_nan: bool = False,
):
    if isinstance(a, na.AbstractVectorArray):
        components_a = a.components
        if isinstance(b, na.AbstractVectorArray):
            if a.type_abstract == b.type_abstract:
                components_b = b.components
            else:
                return NotImplemented
        else:
            components_b = {c: b for c in components_a}
    else:
        if isinstance(b, na.AbstractVectorArray):
            components_b = b.components
            components_a = {c: a for c in components_b}
        else:
            return NotImplemented

    result = True
    for c in components_a:
        result = result * np.allclose(
            a=components_a[c],
            b=components_b[c],
            rtol=rtol,
            atol=atol,
            equal_nan=equal_nan,
        )

    return result


@implements(np.nonzero)
def nonzero(a: na.AbstractVectorArray):
    a = a.explicit
    components = a.components

    components_accumulated = True
    for c in components:
        components_accumulated = components_accumulated & components[c]

    return np.nonzero(components_accumulated)


@implements(np.where)
def where(
    condition: na.AbstractScalar | na.AbstractVectorArray,
    x: float | u.Quantity | na.AbstractScalar | na.AbstractVectorArray,
    y: float | u.Quantity | na.AbstractScalar | na.AbstractVectorArray,
) -> na.AbstractExplicitVectorArray:
    try:
        prototype = vectors._prototype(condition, x, y)
        condition = vectors._normalize(condition, prototype)
        x = vectors._normalize(x, prototype)
        y = vectors._normalize(y, prototype)
    except vectors.VectorTypeError:
        return NotImplemented

    components_condition = condition.components
    components_x = x.components
    components_y = y.components
    components_result = dict()

    for c in components_condition:
        components_result[c] = np.where(
            components_condition[c],
            components_x[c],
            components_y[c],
        )

    return prototype.type_explicit.from_components(components_result)


@implements(np.nan_to_num)
def nan_to_num(
        x: na.AbstractVectorArray,
        copy: bool = True,
        nan: float = 0.0,
        posinf: None | float = None,
        neginf: None | float = None,
):
    components = x.components
    components_result = dict()

    for c in components:
        components_result[c] = np.nan_to_num(
            x=components[c],
            copy=copy,
            nan=nan,
            posinf=posinf,
            neginf=neginf,
        )

    if copy:
        return x.type_explicit.from_components(components_result)
    else:
        if not isinstance(x, na.AbstractExplicitArray):
            raise ValueError("can't write to an array that is not an instance of `named_array.AbstractExplictArray`")
        return x


@implements(np.repeat)
def repeat(
    a: na.AbstractVectorArray,
    repeats: int | na.AbstractScalarArray,
    axis: str,
) -> na.AbstractExplicitVectorArray:

    components = a.broadcasted.components

    components_result = dict()
    for c in components:
        components_result[c] = np.repeat(
            a=components[c],
            repeats=repeats,
            axis=axis,
        )

    return a.type_explicit.from_components(components_result)


@implements(np.diff)
def diff(
    a: na.AbstractVectorArray,
    axis: str,
    n: int = 1,
    prepend: None | float | na.AbstractScalar | na.AbstractVectorArray = None,
    append: None | float | na.AbstractScalar | na.AbstractVectorArray = None,
) -> na.AbstractExplicitVectorArray:

    try:
        prototype = vectors._prototype(a, prepend, append)
        a = vectors._normalize(a, prototype)
        prepend = vectors._normalize(prepend, prototype)
        append = vectors._normalize(append, prototype)
    except vectors.VectorTypeError:     # pragma: nocover
        return NotImplemented

    components_a = a.broadcasted.components
    components_prepend = prepend.components
    components_append = append.components

    result = dict()
    for c in components_a:
        result[c] = np.diff(
            a=components_a[c],
            axis=axis,
            n=n,
            prepend=components_prepend[c],
            append=components_append[c],
        )

    return prototype.type_explicit.from_components(result)


@implements(np.char.mod)
def char_mod(
        a: str | na.AbstractScalar,
        values: str | na.AbstractScalar,
) -> na.ScalarArray:

    try:
        prototype = vectors._prototype(a, values)
        a = vectors._normalize(a, prototype)
        values = vectors._normalize(values, prototype)
    except vectors.VectorTypeError:  # pragma: nocover
        return NotImplemented

    components_a = a.components
    components_values = values.components

    result = dict()
    for c in components_a:
        result[c] = np.char.mod(
            a=components_a[c],
            values=components_values[c],
        )

    return prototype.type_explicit.from_components(result)



from __future__ import annotations
from typing import TypeVar, Generic, Type
import abc
import dataclasses
import numpy as np
import named_arrays as na

__all__ = [
    'AbstractCartesian2dMatrixArray',
    'Cartesian2dMatrixArray',
    'AbstractImplicitCartesian2dMatrixArray',
    'Cartesian2dIdentityMatrixArray',
    'AbstractCartesian2dRotationMatrixArray',
    'Cartesian2dRotationMatrixArray'
]

XT = TypeVar('XT', bound=na.AbstractVectorArray)
YT = TypeVar('YT', bound=na.AbstractVectorArray)


@dataclasses.dataclass(eq=False, repr=False)
class AbstractCartesian2dMatrixArray(
    na.AbstractCartesian2dVectorArray,
    na.AbstractCartesianMatrixArray,
):
    @property
    @abc.abstractmethod
    def x(self) -> na.AbstractVectorArray:
        """
        The `x` row of the matrix.
        """

    @property
    @abc.abstractmethod
    def y(self) -> na.AbstractVectorArray:
        """
        The `y` row of the matrix.
        """

    @property
    def type_abstract(self) -> Type[AbstractCartesian2dMatrixArray]:
        return AbstractCartesian2dMatrixArray

    @property
    def type_explicit(self) -> Type[Cartesian2dMatrixArray]:
        return Cartesian2dMatrixArray



    @property
    def type_vector(self) -> Type[na.Cartesian2dVectorArray]:
        return na.Cartesian2dVectorArray

    @property
    def determinant(self) -> na.ScalarLike:
        if not self.is_square:
            raise ValueError("can only compute determinant of square matrices")
        xx, xy = self.x.components.values()
        yx, yy = self.y.components.values()
        return xx * yy - xy * yx

    def power(
        self,
        exponent: float | na.AbstractScalar,
    ) -> Cartesian2dMatrixArray:
        """
        Compute this matrix raised to the power of a given exponent

        Parameters
        ----------
        exponent
            The power to raise this matrix to.
        """

        z = exponent
        y = z - 1

        a, b = self.x.components.values()
        c, d = self.y.components.values()

        chi = a + d
        delta = a * d - b * c

        t1 = chi / 2
        t2 = np.emath.sqrt(np.square(chi) / 4 - delta)

        e1 = t1 + t2
        e2 = t1 - t2

        d = e1 - e2

        identity = na.Cartesian2dIdentityMatrixArray()

        a1 = (e1 ** z - e2 ** z) / d * self
        a2 = e1 * e2 * (e1 ** y - e2 ** y) / d * identity

        return a1 - a2


@dataclasses.dataclass(eq=False, repr=False)
class Cartesian2dMatrixArray(
    na.Cartesian2dVectorArray,
    AbstractCartesian2dMatrixArray,
    na.AbstractExplicitMatrixArray,
    Generic[XT, YT],
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class AbstractImplicitCartesian2dMatrixArray(
    na.AbstractImplicitCartesian2dVectorArray,
    na.AbstractImplicitMatrixArray,
    AbstractCartesian2dMatrixArray,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class Cartesian2dIdentityMatrixArray(
    AbstractImplicitCartesian2dMatrixArray
):

    @property
    def explicit(self):
        return na.Cartesian2dMatrixArray(
            x=na.Cartesian2dVectorArray(x=1, y=0),
            y=na.Cartesian2dVectorArray(x=0, y=1),
        )

    def power(
        self,
        exponent: float | na.AbstractScalar,
    ) -> AbstractCartesian2dMatrixArray:
        return self


@dataclasses.dataclass(eq=False, repr=False)
class AbstractCartesian2dRotationMatrixArray(
    AbstractImplicitCartesian2dMatrixArray,
):

    @property
    @abc.abstractmethod
    def angle(self) -> na.ScalarLike:
        """
        The angle of the rotation matrix.
        """


@dataclasses.dataclass(eq=False, repr=False)
class Cartesian2dRotationMatrixArray(
    AbstractCartesian2dRotationMatrixArray,
):
    angle: na.ScalarLike = dataclasses.MISSING

    @property
    def explicit(self) -> na.Cartesian2dMatrixArray:
        a = self.angle
        return na.Cartesian2dMatrixArray(
            x=na.Cartesian2dVectorArray(x=np.cos(a), y=-np.sin(a)),
            y=na.Cartesian2dVectorArray(x=np.sin(a), y=np.cos(a)),
        )



import pytest
import numpy as np
import astropy.units as u
import named_arrays as na
from ..cartesian.tests import test_vectors_cartesian

_num_x = test_vectors_cartesian._num_x
_num_y = test_vectors_cartesian._num_y
_num_z = test_vectors_cartesian._num_z
_num_distribution = test_vectors_cartesian._num_distribution


def _spectral_positional_arrays() -> list[na.SpectralPositionalVectorArray]:
    return [
        na.SpectralPositionalVectorArray(
            wavelength=500 * u.nm,
            position=na.Cartesian2dVectorArray(1, 2) * u.mm,
        ),
        na.SpectralPositionalVectorArray(
            wavelength=na.linspace(400, 600, axis="y", num=_num_y) * u.nm,
            position=na.Cartesian2dVectorLinearSpace(1, 2, axis="y", num=_num_y).explicit * u.mm,
        ),
    ]


def _spectral_positional_arrays_2() -> list[na.SpectralPositionalVectorArray]:
    return [
        na.SpectralPositionalVectorArray(
            wavelength=400 * u.nm,
            position=na.Cartesian2dVectorArray(3, 4) * u.m,
        ),
        na.SpectralPositionalVectorArray(
            wavelength=na.NormalUncertainScalarArray(400 * u.nm, width=1 * u.nm),
            position=na.Cartesian2dVectorArray(
                x=na.NormalUncertainScalarArray(3, width=1) * u.m,
                y=na.NormalUncertainScalarArray(4, width=1) * u.m,
            )
        )
    ]


def _spectral_positional_items() -> list[na.AbstractArray | dict[str, int | slice | na.AbstractArray]]:
    return [
        dict(y=0),
        dict(y=slice(0, 1)),
        dict(y=na.ScalarArrayRange(0, 2, axis='y')),
    ]


class AbstractTestAbstractSpectralPositionalVectorArray(
    test_vectors_cartesian.AbstractTestAbstractCartesianVectorArray,
):

    @pytest.mark.parametrize(
        argnames='item',
        argvalues=_spectral_positional_items()
    )
    def test__getitem__(
            self,
            array: na.AbstractSpectralVectorArray,
            item: dict[str, int | slice | na.AbstractArray] | na.AbstractArray
    ):
        super().test__getitem__(array=array, item=item)

    @pytest.mark.parametrize('array_2', _spectral_positional_arrays_2())
    class TestUfuncBinary(
        test_vectors_cartesian.AbstractTestAbstractCartesianVectorArray.TestUfuncBinary
    ):
        pass

    @pytest.mark.parametrize('array_2', _spectral_positional_arrays_2())
    class TestMatmul(
        test_vectors_cartesian.AbstractTestAbstractCartesianVectorArray.TestMatmul,
    ):
        pass

    class TestArrayFunctions(
        test_vectors_cartesian.AbstractTestAbstractCartesianVectorArray.TestArrayFunctions,
    ):

        @pytest.mark.parametrize("array_2", _spectral_positional_arrays_2())
        class TestAsArrayLikeFunctions(
            test_vectors_cartesian.AbstractTestAbstractCartesianVectorArray.TestArrayFunctions.TestAsArrayLikeFunctions,
        ):
            pass

        @pytest.mark.parametrize(
            argnames='where',
            argvalues=[
                np._NoValue,
                True,
                na.ScalarArray(True),
            ]
        )
        class TestReductionFunctions(
            test_vectors_cartesian.AbstractTestAbstractCartesianVectorArray.TestArrayFunctions.TestReductionFunctions,
        ):
            pass

        @pytest.mark.parametrize(
            argnames='q',
            argvalues=[
                .25,
                25 * u.percent,
                na.ScalarLinearSpace(.25, .75, axis='q', num=3, endpoint=True),
            ]
        )
        class TestPercentileLikeFunctions(
            test_vectors_cartesian.AbstractTestAbstractCartesianVectorArray.TestArrayFunctions
            .TestPercentileLikeFunctions,
        ):
            pass

    class TestNamedArrayFunctions(
        test_vectors_cartesian.AbstractTestAbstractCartesianVectorArray.TestNamedArrayFunctions,
    ):
        @pytest.mark.skip
        class TestPltPlotLikeFunctions(
            test_vectors_cartesian.AbstractTestAbstractCartesianVectorArray.TestNamedArrayFunctions
            .TestPltPlotLikeFunctions,
        ):
            pass


@pytest.mark.parametrize("array", _spectral_positional_arrays())
class TestSpectralPositionalVectorArray(
    AbstractTestAbstractSpectralPositionalVectorArray,
    test_vectors_cartesian.AbstractTestAbstractExplicitCartesianVectorArray,
):
    @pytest.mark.parametrize(
        argnames="item",
        argvalues=[
            dict(y=0),
            dict(y=slice(None)),
        ],
    )
    @pytest.mark.parametrize(
        argnames="value",
        argvalues=[
            700 * u.nm,
        ]
    )
    def test__setitem__(
            self,
            array: na.ScalarArray,
            item: dict[str, int | slice | na.ScalarArray] | na.ScalarArray,
            value: float | na.ScalarArray
    ):
        super().test__setitem__(array=array, item=item, value=value)


class AbstractTestAbstractImplicitSpectralPositionalVectorArray(
    AbstractTestAbstractSpectralPositionalVectorArray,
    test_vectors_cartesian.AbstractTestAbstractImplicitCartesianVectorArray,
):
    pass


class AbstractTestAbstractParameterizedSpectralPositionalVectorArray(
    AbstractTestAbstractImplicitSpectralPositionalVectorArray,
    test_vectors_cartesian.AbstractTestAbstractParameterizedCartesianVectorArray,
):
    pass


class AbstractTestAbstractSpectralPositionalVectorSpace(
    AbstractTestAbstractParameterizedSpectralPositionalVectorArray,
    test_vectors_cartesian.AbstractTestAbstractCartesianVectorSpace,
):
    pass


def _spectral_positional_linear_spaces() -> list[na.SpectralPositionalVectorLinearSpace]:
    return [
        na.SpectralPositionalVectorLinearSpace(
            start=400 * u.nm,
            stop=600 * u.nm,
            axis="y",
            num=_num_y,
        )
    ]


@pytest.mark.parametrize("array", _spectral_positional_linear_spaces())
class TestSpectralPositionalVectorLinearSpace(
    AbstractTestAbstractSpectralPositionalVectorSpace,
    test_vectors_cartesian.AbstractTestAbstractCartesianVectorLinearSpace,
):
    pass



from __future__ import annotations
from typing import TYPE_CHECKING, TypeVar, Generic, Sequence, Iterator, Union, Type, Callable, Collection, Any
from typing_extensions import Self
import abc
import dataclasses
import copy
import secrets
import numpy as np
import numpy.typing as npt
import astropy.units as u
import named_arrays as na

__all__ = [
    'QuantityLike',
    'StartT',
    'StopT',
    'named_array_like',
    'get_dtype',
    'value',
    'type_array',
    'broadcast_shapes',
    'shape_broadcasted',
    'ndindex',
    'indices',
    'flatten_axes',
    'axis_normalized',
    'explicit',
    'AbstractArray',
    'ArrayLike',
    'AbstractExplicitArray',
    'AbstractImplicitArray',
    'AbstractRandomMixin',
    'AbstractRangeMixin',
    'AbstractSymmetricRangeMixin',
    'AbstractRandomSample',
    'AbstractParameterizedArray',
    'AbstractLinearParameterizedArrayMixin',
    'AbstractArrayRange',
    'AbstractSpace',
    'AbstractLinearSpace',
    'strata',
    'AbstractStratifiedRandomSpace',
    'AbstractLogarithmicSpace',
    'AbstractGeometricSpace',
    'AbstractUniformRandomSample',
    'AbstractNormalRandomSample',
    'AbstractPoissonRandomSample',
]

QuantityLike = Union[int, float, complex, np.ndarray, u.Quantity]

AxisT = TypeVar("AxisT", bound="str | AbstractArray")
NumT = TypeVar("NumT", bound="int | AbstractArray")
StartT = TypeVar("StartT", bound="QuantityLike | AbstractArray")
StopT = TypeVar("StopT", bound="QuantityLike | AbstractArray")
CenterT = TypeVar("CenterT", bound="QuantityLike | AbstractArray")
WidthT = TypeVar("WidthT", bound="QuantityLike | AbstractArray")
StartExponentT = TypeVar("StartExponentT", bound="QuantityLike | AbstractArray")
StopExponentT = TypeVar("StopExponentT", bound="QuantityLike | AbstractArray")
BaseT = TypeVar("BaseT", bound="QuantityLike | AbstractArray")


def named_array_like(a: Any) -> bool:
    """
    Check if an object is compatible with the :mod:`named_arrays` API.

    If the object has a ``__named_array_function__`` method it is considered compatible.

    Parameters
    ----------
    a
        Object to be checked for compatibility with the :mod:`named_arrays` API.

    Examples
    --------

    Instances of :class:`named_arrays.ScalarArray` are compatible with the :mod:`named_arrays` API

    .. jupyter-execute::

        import named_arrays as na

        na.named_array_like(na.ScalarArray(2))

    But instances of :class:`numpy.ndarray` are not compatible

    .. jupyter-execute::

        import numpy as np

        na.named_array_like(np.empty(3))
    """
    if hasattr(a, "__named_array_function__"):
        return True
    else:
        return False


def get_dtype(
        value: bool | int | float | complex | str | np.ndarray | AbstractArray,
) -> np.dtype:
    """
    Get the equivalent :attr:`numpy.ndarray.dtype` of the argument.

    If the argument is an instance of :class:`numpy.ndarray`, this function simply returns :attr:`numpy.ndarray.dtype`.
    Otherwise, this function wraps the argument in an :func:`numpy.array()` call and then evaluates the ``dtype``.

    Parameters
    ----------
    value
        Object to find the ``dtype`` of

    Returns
    -------
    ``dtype`` of the argument

    """
    if isinstance(value, np.ndarray):
        return value.dtype
    if isinstance(value, AbstractArray):
        if isinstance(value, na.AbstractScalar):
            return value.dtype
        else:
            raise ValueError("non-scalar instances of `na.AbstractArray` may not be represented by a single dtype")
    else:
        return np.array(value).dtype


def value(a: float | u.Quantity | AbstractArray):
    """
    Remove the units (if they exist) from the input.

    Parameters
    ----------
    a
        A numeric value that may or may not have associated units

    Returns
    -------
        The same numeric value, but guaranteed to be dimensionless.
    """

    if isinstance(a, AbstractArray):
        return a.value
    elif isinstance(a, u.Quantity):
        return a.value
    else:
        return a


def type_array(
        *values: bool | int | float | complex | str | np.ndarray | u.Quantity | AbstractArray,
) -> Type[AbstractExplicitArray]:
    cls = None
    priority_max = 0
    for value in values:
        if isinstance(value, AbstractArray):
            cls_tmp = value.type_explicit
            priority_tmp = cls_tmp.__named_array_priority__
            if priority_tmp > priority_max:
                priority_max = priority_tmp
                cls = cls_tmp
    return cls


def broadcast_shapes(*shapes: dict[str, int]) -> dict[str, int]:
    if not shapes:
        return dict()
    result = shapes[0].copy()
    for shape in shapes[1:]:
        if shape == result:
            continue
        for axis in shape:
            shape_axis = shape[axis]
            if axis in result:
                result_axis = result[axis]
                if result_axis == shape_axis:
                    pass
                elif shape_axis == 1:
                    pass
                elif result_axis == 1:
                    result[axis] = shape_axis
                else:
                    raise ValueError(f'shapes {shapes} are not compatible')
            else:
                result[axis] = shape_axis
    return result


def shape_broadcasted(*arrays: Any) -> dict[str, int]:
    shapes = [a.shape for a in arrays if hasattr(a, "__named_array_function__")]
    return broadcast_shapes(*shapes)


def ndindex(
        shape: dict[str, int],
        axis_ignored: None | str | Sequence[str] = None,
) -> Iterator[dict[str, int]]:

    shape = shape.copy()

    if axis_ignored is None:
        axis_ignored = []
    elif isinstance(axis_ignored, str):
        axis_ignored = [axis_ignored]

    for axis in axis_ignored:
        if axis in shape:
            shape.pop(axis)
    shape_tuple = tuple(shape.values())
    for index in np.ndindex(*shape_tuple):
        yield dict(zip(shape.keys(), index))


def indices(shape: dict[str, int]) -> dict[str, na.ScalarArrayRange]:
    return {axis: na.ScalarArrayRange(0, shape[axis], axis=axis) for axis in shape}


def flatten_axes(axes: Sequence[str]):
    if not axes:
        raise ValueError(f"`axes` must be a non-empty sequence, got {axes}")
    return '*'.join(axes)


def axis_normalized(
        a: AbstractArray,
        axis: None | str | Sequence[str],
) -> tuple[str]:
    """
    Convert all the possible values of the ``axis`` argument to a :class:`tuple` of :class:`str`.

    :param a: If ``axis`` is :class:`None` the result is ``a.axes``.
    :param axis: The ``axis`` value to normalize.
    :return: Normalized ``axis`` parameter.
    """

    if axis is None:
        result = a.axes
    elif isinstance(axis, str):
        result = axis,
    else:
        result = tuple(axis)
    return result


def explicit(value: Any | AbstractArray):
    """
    Converts an array to its explicit version if possible.

    Parameters
    ----------
    value
        any value or an instance of :class:`AbstractArray`

    Returns
    -------
        If an instance of :class:`AbstractArray`, returns :attr:`AbstractArray.explicit` otherwise just returns the
        input value.
    """
    if isinstance(value, AbstractArray):
        return value.explicit
    else:
        return value


@dataclasses.dataclass(eq=False, repr=False)
class AbstractArray(
    np.lib.mixins.NDArrayOperatorsMixin,
    abc.ABC,
):
    """
    The ultimate parent class for all array types defined in this package.
    """

    @property
    @abc.abstractmethod
    def __named_array_priority__(self: Self) -> float:
        """
        Attribute used to decide what type of array to return in instances where there is more than one option.

        Similar to :attr:`numpy.class.__array_priority__`

        :return: :type:`int` describing this class's array priority
        """

    @property
    @abc.abstractmethod
    def type_abstract(self: Self) -> Type[AbstractArray]:
        """
        The :class:`AbstractArray` type corresponding to this array
        """

    @property
    @abc.abstractmethod
    def type_explicit(self: Self) -> Type[AbstractExplicitArray]:
        """
        The :class:`AbstractExplicitArray` type corresponding to this array
        """

    @property
    @abc.abstractmethod
    def axes(self: Self) -> tuple[str, ...]:
        """
        A :class:`tuple` of :class:`str` representing the names of each dimension of :attr:`ndarray`.

        Must have the same length as the number of dimensions of :attr:`ndarray`.
        """

    @property
    def axes_flattened(self: Self) -> str:
        """
        Combine :attr:`axes` into a single :class:`str`.

        This is useful for functions like :func:`numpy.flatten` which returns an array with only one dimension.
        """
        return flatten_axes(self.axes)

    @property
    @abc.abstractmethod
    def shape(self: Self) -> dict[str, int]:
        """
        Shape of the array. Analogous to :attr:`numpy.ndarray.shape` but represented as a :class:`dict` where the keys
        are the axis names and the values are the axis sizes.
        """

    @property
    @abc.abstractmethod
    def ndim(self: Self) -> int:
        """
        Number of dimensions of the array. Equivalent to :attr:`numpy.ndarray.ndim`.
        """

    @property
    @abc.abstractmethod
    def size(self: Self) -> int:
        """
        Total number of elements in the array. Equivalent to :attr:`numpy.ndarray.size`
        """

    @property
    @abc.abstractmethod
    def value(self: Self) -> Self:
        """
        Returns a new array with its units removed, if they exist
        """

    @property
    @abc.abstractmethod
    def explicit(self: Self) -> AbstractExplicitArray:
        """
        Converts this array to an instance of :class:`named_arrays.AbstractExplicitArray`
        """

    @property
    def broadcasted(self) -> na.AbstractExplicitArray:
        """
        if this array has multiple components, broadcast them against each other.

        Equivalent to ``a.broadcast_to(a.shape)``
        """
        a = self.explicit
        return a.broadcast_to(a.shape)

    @abc.abstractmethod
    def astype(
            self: Self,
            dtype: str | np.dtype | Type,
            order: str = 'K',
            casting='unsafe',
            subok: bool = True,
            copy: bool = True,
    ) -> Self:
        """
        Copy of the array cast to a specific data type.

        Equivalent to :meth:`numpy.ndarray.astype`.
        """

    @abc.abstractmethod
    def to(
        self: Self,
        unit: u.UnitBase,
        equivalencies: None | list[tuple[u.Unit, u.Unit]] = [],
        copy: bool = True,
    ) -> Self:
        """
        Convert this array to a new unit.

        Equivalent to :meth:`astropy.units.Quantity.to`.

        Parameters
        ----------
        unit
            New unit of the returned array
        equivalencies
            A list of equivalence pairs to try if the units are not directly
            convertible.
        copy
            Boolean flag controlling whether to copy the array.
        """

    @property
    @abc.abstractmethod
    def length(self: Self) -> na.AbstractScalar:
        """
        L2-norm of this array.
        """

    @property
    def indices(self: Self) -> dict[str, na.ScalarArrayRange]:
        return indices(self.shape)

    def ndindex(
            self: Self,
            axis_ignored: None | str | Sequence[str] = None,
    ) -> Iterator[dict[str, int]]:
        return ndindex(
            shape=self.shape,
            axis_ignored=axis_ignored,
        )

    @abc.abstractmethod
    def add_axes(self: Self, axes: str | Sequence[str]) -> AbstractExplicitArray:
        """
        Add new singleton axes to this array

        Parameters
        ----------
        axes
            New axes to add to the array

        Returns
        -------
        Array with new axes added
        """

    @abc.abstractmethod
    def combine_axes(
            self: Self,
            axes: None | Sequence[str] = None,
            axis_new: None | str = None,
    ) -> AbstractExplicitArray:
        """
        Combine some of the axes of the array into a single new axis.

        Parameters
        ----------
        axes
            The axes to combine into a new axis
        axis_new
            The name of the new axis

        Returns
        -------
        Array with the specified axes combined
        """

    @classmethod
    def _lerp(
        cls,
        i: float | na.AbstractScalar,
        a0: float | na.AbstractScalar,
        a1: float | na.AbstractScalar,
    ) -> na.AbstractScalar:
        return a0 * (1 - i) + a1 * i

    def _nlerp(
        self,
        i: dict[str, na.AbstractScalar],
    ) -> na.AbstractExplicitArray:

        if not i:
            return self.explicit

        axis = next(iter(i))

        i_new = {ax: i[ax] for ax in i if ax != axis}

        a0 = self[{axis: slice(None, ~0)}]
        a1 = self[{axis: slice(1, None)}]

        if i_new:
            a0 = a0._nlerp(i_new)
            a1 = a1._nlerp(i_new)

        return self._lerp(i[axis], a0, a1)

    def cell_centers(
        self,
        axis: None | str | Sequence[str] = None,
        random: bool = False,
    ) -> na.AbstractExplicitArray:
        """
        Convert an array from cell vertices to cell centers.

        Parameters
        ----------
        axis
            The axes of the array to average over.
        random
            If true, select a random point within each cell instead of the
            geometric center.
        """

        if axis is None:
            axis = self.axes
        elif isinstance(axis, str):
            axis = (axis, )

        result = self.explicit

        shape = result.shape

        axis = tuple(a for a in axis if a in shape)

        shape_centers = {
            ax: shape[ax] - 1 if ax in axis
            else shape[ax]
            for ax in shape
        }

        if not random:
            i = {
                a: 0.5
                for a in axis
            }
        else:
            i = {
                a: na.random.uniform(0, 1, shape_random=shape_centers)
                for a in axis
            }

        return self._nlerp(i)

    def volume_cell(self, axis: None | str | Sequence[str]) -> na.AbstractScalar:
        """
        Computes the n-dimensional volume of each cell formed by interpreting
        this array as a logically-rectangular grid of vertices.

        Note that this method is usually only used for sorted arrays.

        If `self` is a scalar, this method computes the length of each edge,
        and is equivalent to :func:`numpy.diff`.
        If `self` is a 2d vector, this method computes the area of each
        quadrilateral, and if `self` is a 3d vector, this method computes
        the volume of each cuboid.

        Parameters
        ----------
        axis
            The axis or axes defining the logically-rectangular grid.
            If `self` is a physical scalar, there should only be one axis.
            If `self` is a physical vector, there should be one axis for each
            component of the vector.
        """
        raise NotImplementedError

    def to_string(
            self,
            prefix: None | str = None,
            multiline: None | bool = None,
    ):
        """
        Convert this array instance to a string representation.

        Parameters
        ----------
        prefix
            the length of this string is used to align the output
        multiline
            flag which controls if the output should be spread over multiple
            lines.

        Returns
        -------
        array represented as a :class:`str`
        """
        fields = dataclasses.fields(self)

        if multiline is None:
            multiline_normalized = any(isinstance(getattr(self, f.name), (np.ndarray, na.AbstractArray)) for f in fields)
        else:
            multiline_normalized = multiline

        if multiline_normalized:
            delim_field = "\n"
            pre = " " * len(prefix) if prefix is not None else ""
            tab = " " * 4
        else:
            delim_field = " "
            pre = tab = ""

        result = f"{self.__class__.__qualname__}("
        if multiline_normalized:
            result += "\n"

        for i, f in enumerate(fields):
            field_str = f"{pre}{tab}{f.name}="
            val = getattr(self, f.name)
            if isinstance(val, AbstractArray):
                val_str = val.to_string(prefix=f"{pre}{tab}", multiline=multiline)
            elif isinstance(val, np.ndarray):
                val_str = np.array2string(
                    a=val,
                    max_line_width=120,
                    separator=", ",
                    prefix=field_str,
                )
                if isinstance(val, u.Quantity):
                    val_str = f"{val_str} {val.unit}"
            else:
                val_str = repr(val)
            field_str += val_str
            if multiline_normalized or i < (len(fields) - 1):
                field_str += f",{delim_field}"
            result += field_str
        result += f"{pre})"
        return result

    def __repr__(self):
        return self.to_string()

    @abc.abstractmethod
    def to_string_array(
        self,
        format_value: str = "%.2f",
        format_unit: str = "latex_inline",
        pad_unit: str = r"$\,$",
    ):
        """
        Convert to an array of strings where each string has an
        appropriately-formatted unit attached to the value.

        Parameters
        ----------
        format_value
            The string used to format the numeric value of each element.
        format_unit
            The string used to format the units of each element.
        pad_unit
            The string used to add horizontal space between the value and unit.
        """

    def copy_shallow(self: Self) -> Self:
        return copy.copy(self)

    def copy(self: Self) -> Self:
        return copy.deepcopy(self)

    def __copy__(self: Self) -> Self:
        fields = {field.name: getattr(self, field.name) for field in dataclasses.fields(self)}
        return type(self)(**fields)

    def __deepcopy__(self: Self, memodict={}) -> Self:
        fields = {field.name: copy.deepcopy(getattr(self, field.name)) for field in dataclasses.fields(self)}
        return type(self)(**fields)

    @abc.abstractmethod
    def _getitem(
            self: Self,
            item: dict[str, int | slice | AbstractArray] | AbstractArray,
    ):
        pass

    @abc.abstractmethod
    def _getitem_reversed(
            self: Self,
            array: AbstractArray,
            item: dict[str, int | slice | AbstractArray] | AbstractArray
    ):
        pass

    def __getitem__(
            self: Self,
            item: dict[str, int | slice | AbstractArray] | AbstractArray,
    ) -> AbstractExplicitArray:
        result = self._getitem(item)
        if result is not NotImplemented:
            return result

        else:
            if isinstance(item, dict):
                for ax in item:
                    if isinstance(item[ax], AbstractArray):
                        result = item[ax]._getitem_reversed(self, item)
                        if result is not NotImplemented:
                            return result

            elif isinstance(item, AbstractArray):
                result = item._getitem_reversed(self, item)
                if result is not NotImplemented:
                    return result

        raise ValueError(f"item not supported by array with type {type(self)}")

    @abc.abstractmethod
    def __bool__(self: Self) -> bool:
        return True

    @abc.abstractmethod
    def __mul__(self: Self, other: ArrayLike | u.Unit) -> AbstractExplicitArray:
        return super().__mul__(other)

    @abc.abstractmethod
    def __lshift__(self: Self, other: ArrayLike | u.UnitBase) -> AbstractExplicitArray:
        return super().__lshift__(other)

    @abc.abstractmethod
    def __truediv__(self: Self, other: ArrayLike | u.UnitBase) -> AbstractExplicitArray:
        return super().__truediv__(other)

    @abc.abstractmethod
    def __array_matmul__(
            self: Self,
            x1: ArrayLike,
            x2: ArrayLike,
            out: None | AbstractExplicitArray = None,
            **kwargs,
    ) -> AbstractExplicitArray:
        """
        Method to handle the behavior of :func:`numpy.matmul` which has different behavior than the other ufuncs.
        """
        return NotImplemented

    @abc.abstractmethod
    def __array_ufunc__(
            self,
            function: np.ufunc,
            method: str,
            *inputs,
            **kwargs,
    ) -> None | AbstractArray | tuple[AbstractArray, ...]:
        """
        Method to override the behavior of numpy's ufuncs.
        """
        if function is np.matmul:
            return self.__array_matmul__(*inputs, **kwargs)
        else:
            return NotImplemented

    @abc.abstractmethod
    def __array_function__(
            self: Self,
            func: Callable,
            types: Collection,
            args: tuple,
            kwargs: dict[str, Any],
    ):
        """
        Method to override the behavior of numpy's array functions.
        """
        from . import _core_array_functions

        if func in _core_array_functions.HANDLED_FUNCTIONS:
            return _core_array_functions.HANDLED_FUNCTIONS[func](*args, **kwargs)

        return NotImplemented

    def __named_array_function__(self, func, *args, **kwargs):
        """
        Method used to dispatch custom named array functions
        """
        return NotImplemented

    def broadcast_to(
            self: Self,
            shape: dict[str, int],
    ) -> Self:
        return np.broadcast_to(self, shape=shape)

    def reshape(
            self: Self,
            shape: dict[str, int],
    ) -> Self:
        return np.reshape(self, newshape=shape)

    def min(
            self: Self,
            axis: None | str | Sequence[str] = None,
            initial: npt.ArrayLike = np._NoValue,
            where: Self = np._NoValue,
    ) -> Self:
        return np.min(self, axis=axis, initial=initial, where=where)

    def max(
            self: Self,
            axis: None | str | Sequence[str] = None,
            initial: npt.ArrayLike = np._NoValue,
            where: Self = np._NoValue,
    ) -> Self:
        return np.max(self, axis=axis, initial=initial, where=where)

    def sum(
            self: Self,
            axis: None | str | Sequence[str] = None,
            where: Self = np._NoValue,
    ) -> Self:
        return np.sum(self, axis=axis, where=where)

    def ptp(
            self: Self,
            axis: None | str | Sequence[str] = None,
    ) -> Self:
        return np.ptp(self, axis=axis)

    def mean(
            self: Self,
            axis: None | str | Sequence[str] = None,
            where: Self = np._NoValue,
    ) -> Self:
        return np.mean(self, axis=axis, where=where)

    def std(
            self: Self,
            axis: None | str | Sequence[str] = None,
            where: Self = np._NoValue,
    ) -> Self:
        return np.std(self, axis=axis, where=where)

    def var(
            self: Self,
            axis: None | str | Sequence[str] = None,
            where: Self = np._NoValue,
    ) -> Self:
        return np.var(self, axis=axis, where=where)

    def median(
            self,
            axis: None | str | Sequence[str] = None,
    ):
        return np.median(self, axis=axis)

    def percentile(
            self: Self,
            q: int | float | u.Quantity | Self,
            axis: None | str | Sequence[str] = None,
            out: None | Self = None,
            overwrite_input: bool = False,
            method: str = 'linear',
            keepdims: bool = False,
    ):
        return np.percentile(
            a=self,
            q=q,
            axis=axis,
            out=out,
            overwrite_input=overwrite_input,
            method=method,
            keepdims=keepdims,
        )

    def all(
            self: Self,
            axis: None | str | Sequence[str] = None,
            where: Self = np._NoValue,
    ) -> Self:
        return np.all(self, axis=axis, where=where)

    def any(
            self: Self,
            axis: None | str | Sequence[str] = None,
            where: Self = np._NoValue,
    ) -> Self:
        return np.any(self, axis=axis, where=where)

    def rms(
            self: Self,
            axis: None | str | Sequence[str] = None,
            where: Self = np._NoValue,
    ) -> Self:
        return np.sqrt(np.mean(np.square(self), axis=axis, where=where))

    def vmr(
            self,
            axis: None | str | Sequence[str] = None,
            where: Self = np._NoValue,
    ):
        return na.vmr(self, axis=axis, where=where)

    def transpose(
            self: Self,
            axes: None | Sequence[str] = None,
    ) -> Self:
        return np.transpose(self, axes=axes)

    def _interp_linear_recursive(
            self: Self,
            item: dict[str, Self],
            item_base: dict[str, Self],
    ):
        item = item.copy()

        if not item:
            raise ValueError('Item must contain at least one key')

        axis = next(iter(item))
        x = item.pop(axis)

        if x.shape:
            where_below = x < 0
            where_above = (self.shape[axis] - 1) <= x

            x0 = np.floor(x).astype(int)
            x0[where_below] = 0
            x0[where_above] = self.shape[axis] - 2

        else:
            if x < 0:
                x0 = 0
            elif x >= self.shape[axis] - 1:
                x0 = self.shape[axis] - 2
            else:
                x0 = int(x)

        x1 = x0 + 1

        item_base_0 = {**item_base, axis: x0}
        item_base_1 = {**item_base, axis: x1}

        if item:
            y0 = self._interp_linear_recursive(item=item, item_base=item_base_0, )
            y1 = self._interp_linear_recursive(item=item, item_base=item_base_1, )
        else:
            y0 = self[item_base_0].astype(float)
            y1 = self[item_base_1].astype(float)

        result = y0 + (x - x0) * (y1 - y0)
        return result

    def interp_linear(
            self: Self,
            item: dict[str, Self],
    ) -> Self:
        if item:
            return self._interp_linear_recursive(
                item=item,
                item_base=self[{ax: 0 for ax in item}].indices,
            )
        else:
            return self

    def __call__(self: Self, item: dict[str, Self]) -> Self:
        return self.interp_linear(item=item)

    def index_secant(
            self: Self,
            value: Self,
            axis: None | str | Sequence[str] = None,
    ) -> dict[str, Self]:

        import named_arrays.scalars
        import named_arrays.vectors
        import named_arrays.optimization

        if axis is None:
            axis = list(self.shape.keys())
        elif isinstance(axis, str):
            axis = [axis, ]

        shape = self.shape
        shape_nearest = named_arrays.vectors.CartesianND({ax: shape[ax] for ax in axis})

        if isinstance(self, named_arrays.vectors.VectorInterface):
            coordinates = self.coordinates
            coordinates = {comp: None if value.coordinates[comp] is None else coordinates[comp] for comp in coordinates}
            self_subspace = type(self).from_coordinates(coordinates)
        else:
            self_subspace = self

        def indices_factory(index: named_arrays.vectors.CartesianND) -> dict[str, named_arrays.scalars.Scalar]:
            return index.coordinates

        def get_index(index: named_arrays.vectors.CartesianND) -> named_arrays.vectors.CartesianND:
            index = indices_factory(index)
            print(self_subspace)
            value_new = self_subspace(index)
            diff = value_new - value
            if isinstance(diff, named_arrays.vectors.AbstractVector):
                diff = named_arrays.vectors.CartesianND({c: diff.coordinates[c] for c in diff.coordinates if diff.coordinates[c] is not None})
            return diff

        result = named_arrays.optimization.root_finding.secant(
            func=get_index,
            root_guess=shape_nearest // 2,
            step_size=named_arrays.vectors.CartesianND({ax: 1e-6 for ax in axis}),
        )

        return indices_factory(result)

    def index(
            self: Self,
            value: Self,
            axis: None | str | Sequence[str] = None,
    ) -> dict[str, Self]:
        return self.index_secant(value=value, axis=axis)


ArrayLike = Union[QuantityLike, AbstractArray]


@dataclasses.dataclass(eq=False, repr=False)
class AbstractExplicitArray(
    AbstractArray,
):
    @classmethod
    @abc.abstractmethod
    def from_scalar_array(
            cls: type[Self],
            a: float | u.Quantity | na.AbstractScalarArray,
            like: None | Self = None,
    ) -> Self:
        """
        Constructs a new version of this array using ``a`` as the underlying data.

        Parameters
        ----------
        a
            Anything that can be coerced into an instance of :class:`named_arrays.AbstractScalarArray`.
        like
            Optional reference object.
            If provided, the result will be defined by this object.
        """
        if like is None:
            return cls()
        else:
            if isinstance(like, cls):
                return type(like)()
            else:
                raise TypeError(
                    f"If `like` is not `None`, it must be an instance of `{cls.__name__}`, "
                    f"got `{type(like).__name__}`"
                )

    @abc.abstractmethod
    def __setitem__(
            self,
            item,
            value,
    ) -> None:
        pass


@dataclasses.dataclass(eq=False, repr=False)
class AbstractImplicitArray(
    AbstractArray,
):
    @property
    def axes(self: Self) -> tuple[str, ...]:
        return self.explicit.axes

    @property
    def ndim(self: Self) -> int:
        return self.explicit.ndim

    @property
    def size(self: Self) -> int:
        return self.explicit.size

    @property
    def value(self: Self) -> Self:
        return self.explicit.value

    @property
    def shape(self: Self) -> dict[str, int]:
        return self.explicit.shape

    @abc.abstractmethod
    def _attr_normalized(self, name: str) -> AbstractExplicitArray:
        """
        Similar to :func:`getattr`, but normalizes it to an instance of :class:`AbstractExplicitArray`
        """


@dataclasses.dataclass(eq=False, repr=False)
class AbstractRandomMixin(
    abc.ABC,
):

    def __post_init__(self):
        if self.seed is None:
            self.seed = secrets.randbits(128)

    @property
    @abc.abstractmethod
    def seed(self: Self) -> int:
        """
        Seed for the random number generator instance
        """

    @seed.setter
    @abc.abstractmethod
    def seed(self: Self, value: int) -> None:
        pass

    @property
    def _rng(self: Self) -> np.random.Generator:
        return np.random.default_rng(seed=self.seed)


@dataclasses.dataclass(eq=False, repr=False)
class AbstractRangeMixin(
    abc.ABC,
):

    @property
    @abc.abstractmethod
    def start(self: Self) -> int | AbstractArray:
        """
        Starting value of the range.
        """

    @property
    @abc.abstractmethod
    def stop(self: Self) -> int | AbstractArray:
        """
        Ending value of the range.
        """

    @property
    def range(self: Self) -> AbstractArray:
        return self.stop - self.start


@dataclasses.dataclass(eq=False, repr=False)
class AbstractSymmetricRangeMixin(
    AbstractRangeMixin,
):
    @property
    @abc.abstractmethod
    def center(self: Self) -> ArrayLike:
        """
        Center value of the range.
        """

    @property
    @abc.abstractmethod
    def width(self: Self) -> ArrayLike:
        """
        Width of the range.
        """

    @property
    def start(self: Self) -> ArrayLike:
        return self.center - self.width

    @property
    def stop(self: Self) -> ArrayLike:
        return self.center + self.width


@dataclasses.dataclass(eq=False, repr=False)
class AbstractRandomSample(
    AbstractRandomMixin,
    AbstractImplicitArray,
):

    @property
    @abc.abstractmethod
    def shape_random(self: Self) -> None | dict[str, int]:
        """
        Dimensions along which the resulting random sample is completely uncorrelated.
        """


@dataclasses.dataclass(eq=False, repr=False)
class AbstractUniformRandomSample(
    AbstractRangeMixin,
    AbstractRandomSample,
    Generic[StartT, StopT],
):
    start: StartT = dataclasses.MISSING
    stop: StopT = dataclasses.MISSING
    shape_random: None | dict[str, int] = None
    seed: None | int = None

    @property
    def explicit(self) -> AbstractExplicitArray:

        start = self._attr_normalized("start")
        stop = self._attr_normalized("stop")

        return na.random.uniform(
            low=start,
            high=stop,
            shape_random=self.shape_random,
            seed=self.seed,
        )


@dataclasses.dataclass
class AbstractNormalRandomSample(
    AbstractSymmetricRangeMixin,
    AbstractRandomSample,
    Generic[CenterT, WidthT],
):
    center: CenterT = dataclasses.MISSING
    width: WidthT = dataclasses.MISSING
    shape_random: None | dict[str, int] = None
    seed: None | int = None

    @property
    def explicit(self) -> AbstractExplicitArray:
        center = self._attr_normalized("center")
        width = self._attr_normalized("width")

        return na.random.normal(
            loc=center,
            scale=width,
            shape_random=self.shape_random,
            seed=self.seed,
        )


@dataclasses.dataclass
class AbstractPoissonRandomSample(
    AbstractRandomSample,
    Generic[CenterT],
):
    center: CenterT = dataclasses.MISSING
    shape_random: None | dict[str, int] = None
    seed: None | int = None

    @property
    def explicit(self) -> AbstractExplicitArray:
        center = self._attr_normalized("center")

        return na.random.poisson(
            lam=center,
            shape_random=self.shape_random,
            seed=self.seed,
        )


@dataclasses.dataclass(eq=False, repr=False)
class AbstractParameterizedArray(
    AbstractImplicitArray,
):

    @property
    @abc.abstractmethod
    def axis(self: Self) -> str | AbstractArray:
        """
        The axis along which the array is parameterized
        """

    @property
    @abc.abstractmethod
    def num(self: Self) -> int | AbstractArray:
        """
        Number of elements in the parameterization
        """


@dataclasses.dataclass(eq=False, repr=False)
class AbstractLinearParameterizedArrayMixin(
    abc.ABC
):
    @property
    @abc.abstractmethod
    def step(self: Self) -> int | AbstractArray:
        """
        Spacing between the values.
        """


@dataclasses.dataclass(eq=False, repr=False)
class AbstractArrayRange(
    AbstractLinearParameterizedArrayMixin,
    AbstractRangeMixin,
    AbstractParameterizedArray,
    Generic[StartT, StopT]
):
    start: StartT = dataclasses.MISSING
    stop: StopT = dataclasses.MISSING
    axis: str | na.AbstractArray= dataclasses.MISSING
    step: int | float | na.AbstractArray = 1

    @property
    def explicit(self: Self) -> AbstractExplicitArray:
        start = self._attr_normalized("start")
        stop = self._attr_normalized("stop")

        return na.arange(
            start=start,
            stop=stop,
            axis=self.axis,
            step=self.step,
        )

    @property
    def num(self: Self) -> int:
        return np.ceil((self.stop - self.start) / self.step).astype(int)


@dataclasses.dataclass(eq=False, repr=False)
class AbstractSpace(
    AbstractParameterizedArray,
):
    @property
    @abc.abstractmethod
    def endpoint(self: Self) -> bool:
        """
        If ``True``, :attr:`stop` is the last sample, otherwise it is not included.
        """


@dataclasses.dataclass(eq=False, repr=False)
class AbstractLinearSpace(
    AbstractLinearParameterizedArrayMixin,
    AbstractRangeMixin,
    AbstractSpace,
    Generic[StartT, StopT, AxisT, NumT],
):
    start: StartT = dataclasses.MISSING
    stop: StopT = dataclasses.MISSING
    axis: AxisT = dataclasses.MISSING
    num: NumT = 11
    endpoint: bool = True
    centers: bool = False

    @property
    def explicit(self: Self) -> AbstractExplicitArray:
        return na.linspace(
            start=self._attr_normalized("start"),
            stop=self._attr_normalized("stop"),
            axis=self.axis,
            num=self.num,
            endpoint=self.endpoint,
            centers=self.centers,
        )

    @property
    def step(self: Self) -> AbstractArray:
        return na.step(
            start=self.start,
            stop=self.stop,
            num=self.num,
            endpoint=self.endpoint,
            centers=self.centers,
        )


def strata(a: AbstractArray) -> AbstractArray:
    """
    If ``a`` is an instance of :class:`AbstractStratifiedRandomSpace`,
    return ``a.strata``, otherwise return ``a``

    Parameters
    ----------
    a
        An array to isolate the strata of.

    Examples
    --------

    Make a scatterplot of a 2D stratified random array and the centers of
    the strata.

    .. jupyter-execute::

        import matplotlib.pyplot as plt
        import named_arrays as na

        # Define a 2D stratified random array.
        a = na.Cartesian2dVectorStratifiedRandomSpace(
            start=-1,
            stop=1,
            axis=na.Cartesian2dVectorArray("x", "y"),
            num=11,
        )

        # Isolate the strata of the array
        strata = na.strata(a)

        # Plot the array and the strata
        fig, ax = plt.subplots()
        na.plt.scatter(
            a.x,
            a.y,
            ax=ax,
        )
        na.plt.scatter(
            strata.x,
            strata.y,
            ax=ax,
        );
    """
    if isinstance(a, AbstractStratifiedRandomSpace):
        return a.strata
    else:
        return a


@dataclasses.dataclass(eq=False, repr=False)
class AbstractStratifiedRandomSpace(
    AbstractRandomMixin,
    AbstractLinearSpace[StartT, StopT, AxisT, NumT],
):
    centers: bool = True
    seed: None | int = None

    @property
    def explicit(self: Self) -> AbstractExplicitArray:
        result = self.strata

        step_size = self.step

        delta = na.random.uniform(
            low=-step_size / 2,
            high=step_size / 2,
            shape_random=result.shape,
            seed=self.seed,
        )

        return result + delta

    @property
    def strata(self: Self) -> AbstractExplicitArray:
        return na.linspace(
            start=self._attr_normalized("start"),
            stop=self._attr_normalized("stop"),
            num=self.num,
            endpoint=self.endpoint,
            axis=self.axis,
            centers=self.centers,
        )


@dataclasses.dataclass(eq=False, repr=False)
class AbstractLogarithmicSpace(
    AbstractRangeMixin,
    AbstractSpace,
    Generic[StartExponentT, StopExponentT, BaseT, AxisT, NumT]
):
    start_exponent: StartExponentT = dataclasses.MISSING
    stop_exponent: StopExponentT = dataclasses.MISSING
    base: BaseT = dataclasses.MISSING
    axis: AxisT = dataclasses.MISSING
    num: NumT = 11
    endpoint: bool = True

    @property
    def explicit(self: Self) -> AbstractExplicitArray:
        return na.logspace(
            start=self._attr_normalized("start_exponent"),
            stop=self._attr_normalized("stop_exponent"),
            axis=self.axis,
            num=self.num,
            endpoint=self.endpoint,
            base=self.base,
        )

    @property
    def start(self: Self) -> ArrayLike:
        return self.base ** self.start_exponent

    @property
    def stop(self: Self) -> ArrayLike:
        return self.base ** self.stop_exponent


@dataclasses.dataclass(eq=False, repr=False)
class AbstractGeometricSpace(
    AbstractRangeMixin,
    AbstractSpace,
    Generic[StartT, StopT, AxisT, NumT]
):
    start: StartT = dataclasses.MISSING
    stop: StopT = dataclasses.MISSING
    axis: AxisT = dataclasses.MISSING
    num: NumT = 11
    endpoint: bool = True

    @property
    def explicit(self: Self) -> AbstractExplicitArray:
        return na.geomspace(
            start=self._attr_normalized("start"),
            stop=self._attr_normalized("stop"),
            axis=self.axis,
            num=self.num,
            endpoint=self.endpoint,
        )



from typing import Callable, Sequence, Literal
import collections
import numpy as np
import numpy.typing as npt
import matplotlib.axes
import matplotlib.artist
import matplotlib.pyplot as plt
import astropy.units as u
import named_arrays as na
import named_arrays._scalars.scalar_named_array_functions
from . import uncertainties

__all__ = [
    "ASARRAY_LIKE_FUNCTIONS",
    "RANDOM_FUNCTIONS",
    "PLT_PLOT_LIKE_FUNCTIONS",
    "HANDLED_FUNCTIONS",
    "random",
]

ASARRAY_LIKE_FUNCTIONS = named_arrays._scalars.scalar_named_array_functions.ASARRAY_LIKE_FUNCTIONS
RANDOM_FUNCTIONS = (
    na.random.uniform,
    na.random.normal,
    na.random.poisson,
    na.random.binomial,
    na.random.gamma,
)
PLT_PLOT_LIKE_FUNCTIONS = named_arrays._scalars.scalar_named_array_functions.PLT_PLOT_LIKE_FUNCTIONS
NDFILTER_FUNCTIONS = named_arrays._scalars.scalar_named_array_functions.NDFILTER_FUNCTIONS
HANDLED_FUNCTIONS = dict()


def _implements(function: Callable):
    """Register a __named_array_function__ implementation for AbstractScalarArray objects."""
    def decorator(func):
        HANDLED_FUNCTIONS[function] = func
        return func
    return decorator


def asarray_like(
        func: Callable,
        a: None | float | u.Quantity | na.AbstractScalarArray | na.AbstractUncertainScalarArray,
        dtype: None | type | np.dtype = None,
        order: None | str = None,
        *,
        like: None | na.AbstractUncertainScalarArray = None,
) -> None | na.UncertainScalarArray:

    if isinstance(a, na.AbstractArray):
        if isinstance(a, na.AbstractUncertainScalarArray):
            a_nominal = a.nominal
            a_distribution = a.distribution
        elif isinstance(a, na.AbstractScalarArray):
            a_nominal = a_distribution = a
        else:
            return NotImplemented
    else:
        a_nominal = a_distribution = a

    if isinstance(like, na.AbstractArray):
        if isinstance(like, na.AbstractUncertainScalarArray):
            like_nominal = like.nominal
            like_distribution = like.distribution
            type_like = like.type_explicit
        elif isinstance(like, na.AbstractScalarArray):
            like_nominal = like_distribution = like
            type_like = na.UncertainScalarArray
        else:
            return NotImplemented
    else:
        like_nominal = like_distribution = like
        type_like = na.UncertainScalarArray

    return type_like(
        nominal=func(
            a=a_nominal,
            dtype=dtype,
            order=order,
            like=like_nominal,
        ),
        distribution=func(
            a=a_distribution,
            dtype=dtype,
            order=order,
            like=like_distribution,
        ),
    )


@_implements(na.unit)
def unit(
        a: na.AbstractUncertainScalarArray,
        unit_dimensionless: None | float | u.UnitBase = None,
        squeeze: bool = True,
) -> None | u.UnitBase:
    return na.unit(
        a=a.nominal,
        unit_dimensionless=unit_dimensionless,
    )


@_implements(na.unit_normalized)
def unit_normalized(
        a: na.AbstractUncertainScalarArray,
        unit_dimensionless: float | u.UnitBase = u.dimensionless_unscaled,
        squeeze: bool = True,
) -> u.UnitBase:
    return na.unit_normalized(
        a.nominal,
        unit_dimensionless=unit_dimensionless,
    )


@_implements(na.interp)
def interp(
        x: float | u.Quantity | na.AbstractScalar,
        xp:  na.AbstractScalar,
        fp: na.AbstractScalar,
        axis: None | str = None,
        left: None | float | u.Quantity | na.AbstractScalar = None,
        right: None | float | u.Quantity | na.AbstractScalar = None,
        period: None | float | u.Quantity | na.AbstractScalar = None,
) -> na.AbstractUncertainScalarArray:
    try:
        x = uncertainties._normalize(x)
        xp = uncertainties._normalize(xp)
        fp = uncertainties._normalize(fp)
        left = uncertainties._normalize(left)
        right = uncertainties._normalize(right)
        period = uncertainties._normalize(period)
    except na.UncertainScalarTypeError:
        return NotImplemented

    result = x.type_explicit(
        nominal=na.interp(
            x=x.nominal,
            xp=xp.nominal,
            fp=fp.nominal,
            left=left.nominal,
            right=right.nominal,
            period=period.nominal,
        ),
        distribution=na.interp(
            x=x.distribution,
            xp=xp.distribution,
            fp=fp.distribution,
            left=left.distribution,
            right=right.distribution,
            period=period.distribution,
        ),
    )

    return result


@_implements(na.histogram)
def histogram(
    a: na.AbstractScalar,
    bins: dict[str, int] | na.AbstractScalar,
    axis: None | str | Sequence[str] = None,
    min: None | na.AbstractScalar = None,
    max: None | na.AbstractScalar = None,
    density: bool = False,
    weights: None | na.AbstractScalar = None,
) -> na.FunctionArray[na.AbstractScalar, na.AbstractScalar]:

    if isinstance(bins, na.AbstractArray):
        bins = (bins, )

    hist, edges = na.histogramdd(
        a,
        bins=bins,
        axis=axis,
        min=min,
        max=max,
        density=density,
        weights=weights,
    )
    edges = edges[0]

    return na.FunctionArray(
        inputs=edges,
        outputs=hist,
    )


@_implements(na.histogram2d)
def histogram2d(
    x: na.AbstractScalar,
    y: na.AbstractScalar,
    bins: dict[str, int] | na.AbstractCartesian2dVectorArray,
    axis: None | str | Sequence[str] = None,
    min: None | na.AbstractScalar | na.AbstractCartesian2dVectorArray = None,
    max: None | na.AbstractScalar | na.AbstractCartesian2dVectorArray = None,
    density: bool = False,
    weights: None | na.AbstractScalar = None,
) -> na.FunctionArray[na.Cartesian2dVectorArray, na.UncertainScalarArray]:
    try:
        x = uncertainties._normalize(x).broadcasted
        y = uncertainties._normalize(y).broadcasted
        weights = uncertainties._normalize(weights)

        if isinstance(bins, na.AbstractCartesian2dVectorArray):
            bins = na.Cartesian2dVectorArray(
                x=uncertainties._normalize(bins.x),
                y=uncertainties._normalize(bins.y),
            )
            bins_nominal = na.Cartesian2dVectorArray(
                x=bins.x.nominal,
                y=bins.y.nominal,
            )
            bins_distribution = na.Cartesian2dVectorArray(
                x=bins.x.distribution,
                y=bins.y.distribution,
            )
        else:
            bins_nominal = bins_distribution = bins

        if min is not None:
            if not isinstance(min, na.AbstractCartesian2dVectorArray):
                min = na.Cartesian2dVectorArray.from_scalar(min)
            min = na.Cartesian2dVectorArray(
                x=uncertainties._normalize(min.x),
                y=uncertainties._normalize(min.y),
            )
            min_nominal = na.Cartesian2dVectorArray(
                x=min.x.nominal,
                y=min.y.nominal,
            )
            min_distribution = na.Cartesian2dVectorArray(
                x=min.x.distribution,
                y=min.y.distribution,
            )
        else:
            min_nominal = min_distribution = min

        if max is not None:
            if not isinstance(max, na.AbstractCartesian2dVectorArray):
                max = na.Cartesian2dVectorArray.from_scalar(max)
            max = na.Cartesian2dVectorArray(
                x=uncertainties._normalize(max.x),
                y=uncertainties._normalize(max.y),
            )
            max_nominal = na.Cartesian2dVectorArray(
                x=max.x.nominal,
                y=max.y.nominal,
            )
            max_distribution = na.Cartesian2dVectorArray(
                x=max.x.distribution,
                y=max.y.distribution,
            )
        else:
            max_nominal = max_distribution = max

    except na.UncertainScalarTypeError:  # pragma: nocover
        return NotImplemented

    result_nominal = na.histogram2d(
        x=x.nominal,
        y=y.nominal,
        bins=bins_nominal,
        axis=axis,
        min=min_nominal,
        max=max_nominal,
        density=density,
        weights=weights.nominal,
    )

    result_distribution = na.histogram2d(
        x=x.distribution,
        y=y.distribution,
        bins=bins_distribution,
        axis=axis,
        min=min_distribution,
        max=max_distribution,
        density=density,
        weights=weights.distribution,
    )

    return na.FunctionArray(
        inputs=na.Cartesian2dVectorArray(
            x=na.UncertainScalarArray(
                nominal=result_nominal.inputs.x,
                distribution=result_distribution.inputs.x,
            ),
            y=na.UncertainScalarArray(
                nominal=result_nominal.inputs.y,
                distribution=result_distribution.inputs.y,
            ),
        ),
        outputs=na.UncertainScalarArray(
            nominal=result_nominal.outputs,
            distribution=result_distribution.outputs,
        ),
    )


@_implements(na.histogramdd)
def histogramdd(
    *sample: na.AbstractScalar,
    bins: dict[str, int] | na.AbstractScalar | Sequence[na.AbstractScalar],
    axis: None | str | Sequence[str] = None,
    min: None | na.AbstractScalar | Sequence[na.AbstractScalar] = None,
    max: None | na.AbstractScalar | Sequence[na.AbstractScalar] = None,
    density: bool = False,
    weights: None | na.AbstractScalar = None,
) -> tuple[na.AbstractScalar, tuple[na.AbstractScalar, ...]]:

    try:
        sample = [uncertainties._normalize(s) for s in sample]
        bins = [uncertainties._normalize(b) for b in bins] if not isinstance(bins, dict) else bins
        weights = uncertainties._normalize(weights) if weights is not None else weights
    except uncertainties.ScalarTypeError:  # pragma: nocover
        return NotImplemented

    shape = na.shape_broadcasted(*sample, weights)

    if axis is None:
        axis = tuple(shape)
    elif isinstance(axis, str):
        axis = (axis,)

    shape_hist = {ax: shape[ax] for ax in axis}

    shape_sample = na.shape_broadcasted(*sample)
    shape_sample = na.broadcast_shapes(shape_sample, shape_hist)
    sample = [s.broadcast_to(shape_sample) for s in sample]

    if weights is not None:
        shape_weights = na.shape(weights)
        shape_weights = na.broadcast_shapes(shape_weights, shape_hist)
        weights = weights.broadcast_to(shape_weights)

    if min is None:
        min = [s.min(axis) for s in sample]
    elif not isinstance(min, collections.abc.Sequence):
        min = [min] * len(sample)

    if max is None:
        max = [s.max(axis) for s in sample]
    elif not isinstance(max, collections.abc.Sequence):
        max = [max] * len(sample)

    try:
        max = [uncertainties._normalize(m) for m in max]
        min = [uncertainties._normalize(m) for m in min]
    except uncertainties.ScalarTypeError:  # pragma: nocover
        return NotImplemented

    hist_nominal, edges_nominal = na.histogramdd(
        *[na.as_named_array(s.nominal) for s in sample],
        bins=[b.nominal for b in bins] if not isinstance(bins, dict) else bins,
        axis=axis,
        min=[m.nominal for m in min],
        max=[m.nominal for m in max],
        density=density,
        weights=weights.nominal if weights is not None else weights,
    )

    hist_distribution, edges_distribution = na.histogramdd(
        *[na.as_named_array(s.distribution) for s in sample],
        bins=[b.distribution for b in bins] if not isinstance(bins, dict) else bins,
        axis=axis,
        min=[m.distribution for m in min],
        max=[m.distribution for m in max],
        density=density,
        weights=weights.distribution if weights is not None else weights,
    )

    hist = na.UncertainScalarArray(
        nominal=hist_nominal,
        distribution=hist_distribution,
    )

    edges = [
        na.UncertainScalarArray(e_n, e_d)
        for e_n, e_d in zip(edges_nominal, edges_distribution)
    ]

    return hist, edges


def random(
        func: Callable,
        *args: float | u.Quantity | na.AbstractScalar,
        shape_random: None | dict[str, int] = None,
        seed: None | int = None,
        **kwargs: float | u.Quantity | na.AbstractScalar,
) -> na.UncertainScalarArray:

    try:
        args = tuple(uncertainties._normalize(arg) for arg in args)
        kwargs = {k: uncertainties._normalize(kwargs[k]) for k in kwargs}
    except na.UncertainScalarTypeError:
        return NotImplemented

    return na.UncertainScalarArray(
        nominal=func(
            *tuple(arg.nominal for arg in args),
            shape_random=shape_random,
            seed=seed,
            **{k: kwargs[k].nominal for k in kwargs},
        ),
        distribution=func(
            *tuple(arg.distribution for arg in args),
            shape_random=shape_random,
            seed=seed,
            **{k: kwargs[k].distribution for k in kwargs},
        )
    )


def plt_plot_like(
        func: Callable,
        *args: na.AbstractScalar,
        ax: None | matplotlib.axes.Axes = None,
        axis: None | str = None,
        where: bool | na.AbstractScalarArray = True,
        components: None | tuple[str, ...] = None,
        **kwargs,
) -> na.UncertainScalarArray[
    npt.NDArray[matplotlib.artist.Artist],
    npt.NDArray[matplotlib.artist.Artist]
]:

    if components is not None:
        raise ValueError(f"`components` should be `None` for scalars, got {components}")

    try:
        args = tuple(uncertainties._normalize(arg) for arg in args)
        where = uncertainties._normalize(where)
        kwargs = {k: uncertainties._normalize(kwargs[k]) for k in kwargs}
    except na.UncertainScalarTypeError:
        return NotImplemented

    shape_args = na.shape_broadcasted(*args)

    shape = na.broadcast_shapes(na.shape(ax), shape_args)

    if axis is None:
        if len(shape_args) != 1:
            raise ValueError(
                f"if `axis` is `None`, the broadcasted shape of `*args`, "
                f"{shape_args}, should have one element"
            )
        axis = next(iter(shape_args))

    args = tuple(arg.broadcast_to(shape) for arg in args)

    axis_distribution = args[0].axis_distribution
    shape_distribution = na.broadcast_shapes(*[arg.shape_distribution for arg in args])

    if axis_distribution in shape_distribution:
        num_distribution = shape_distribution[axis_distribution]
    else:
        num_distribution = 1

    if num_distribution == 0:
        alpha = 1
    else:
        alpha = max(1 / num_distribution, 1/255)
    if "alpha" in kwargs:
        kwargs["alpha"] *= alpha
    else:
        kwargs["alpha"] = na.UncertainScalarArray(1, alpha)

    if "color" not in kwargs:
        color_cycle = plt.rcParams['axes.prop_cycle'].by_key()['color']
        shape_orthogonal = {a: shape[a] for a in shape if a != axis}
        color = na.ScalarArray.empty(shape=shape_orthogonal, dtype=object)
        for i, index in enumerate(color.ndindex()):
            color[index] = color_cycle[i % len(color_cycle)]
        kwargs["color"] = uncertainties._normalize(color)

    result = na.UncertainScalarArray(
        nominal=func(
            *tuple(na.as_named_array(arg.nominal) for arg in args),
            ax=ax,
            axis=axis,
            where=where.nominal,
            **{k: kwargs[k].nominal for k in kwargs}
        ),
        distribution=func(
            *tuple(na.as_named_array(arg.distribution) for arg in args),
            ax=ax,
            axis=axis,
            where=where.distribution,
            **{k: kwargs[k].distribution for k in kwargs}
        )
    )

    return result


@_implements(na.plt.scatter)
def plt_scatter(
        *args: na.AbstractScalar,
        s: None | na.AbstractScalar = None,
        c: None | na.AbstractScalar = None,
        ax: None | matplotlib.axes.Axes | na.ScalarArray = None,
        where: bool | na.AbstractScalar = True,
        components: None | tuple[str, ...] = None,
        **kwargs,
) -> na.UncertainScalarArray:

    if components is not None:
        raise ValueError(
            f"`components` should be `None` for scalars, got {components}"
        )

    try:
        args = tuple(uncertainties._normalize(arg) for arg in args)
        s = uncertainties._normalize(s)
        c = uncertainties._normalize(c)
        where = uncertainties._normalize(where)
        kwargs = {k: uncertainties._normalize(kwargs[k]) for k in kwargs}
    except na.UncertainScalarTypeError:
        return NotImplemented

    if ax is None:
        ax = plt.gca()
    ax = na.as_named_array(ax)

    axis_distribution = args[0].axis_distribution
    shape_distribution = na.broadcast_shapes(*[arg.shape_distribution for arg in args])

    if axis_distribution in shape_distribution:
        num_distribution = shape_distribution[axis_distribution]
    else:
        num_distribution = 1

    if num_distribution == 0:
        alpha = 1
    else:
        alpha = max(1 / num_distribution, 1 / 255)

    if "alpha" in kwargs:
        kwargs["alpha"] = kwargs["alpha"] * alpha
    else:
        kwargs["alpha"] = na.UncertainScalarArray(1, alpha)

    result_nominal = na.plt.scatter(
        *[na.as_named_array(arg.nominal) for arg in args],
        s=s.nominal,
        c=c.nominal,
        ax=ax,
        where=where.nominal,
        components=components,
        **{k: kwargs[k].nominal for k in kwargs},
    )

    if c.distribution is None:
        c_distribution = na.ScalarArray.zeros(shape=ax.shape | dict(rgba=4))
        for index in ax.ndindex():
            facecolor = result_nominal[index].ndarray.get_facecolor()[0]
            c_distribution[index] = na.ScalarArray(
                ndarray=facecolor,
                axes="rgba",
            )
        c.distribution = c_distribution

    result_distribution = na.plt.scatter(
        *[na.as_named_array(arg.distribution) for arg in args],
        s=s.distribution,
        c=c.distribution,
        ax=ax,
        where=where.distribution,
        components=components,
        **{k: kwargs[k].distribution for k in kwargs},
    )

    result = na.UncertainScalarArray(
        nominal=result_nominal,
        distribution=result_distribution,
    )

    return result


@_implements(na.plt.stairs)
def plt_stairs(
        *args: na.AbstractScalar,
        ax: None | matplotlib.axes.Axes = None,
        axis: None | str = None,
        where: bool | na.AbstractScalarArray = True,
        **kwargs,
) -> na.UncertainScalarArray[
    npt.NDArray[matplotlib.artist.Artist],
    npt.NDArray[matplotlib.artist.Artist]
]:

    if len(args) == 1:
        edges = None
        values, = args
    elif len(args) == 2:
        edges, values = args
    else:   # pragma: nocover
        raise ValueError(
            f"incorrect number of arguments, expected 1 or 2, got {len(args)}"
        )

    try:
        values = uncertainties._normalize(values)
        edges = uncertainties._normalize(edges) if edges is not None else edges
        where = uncertainties._normalize(where)
        kwargs = {k: uncertainties._normalize(kwargs[k]) for k in kwargs}
    except na.UncertainScalarTypeError:     # pragma: nocover
        return NotImplemented

    if axis is None:
        if len(values.shape) != 1:
            raise ValueError(
                f"if {axis=}, {values.shape=} should have only one element."
            )
        axis = next(iter(values.shape))
    else:
        if axis not in values.shape:
            raise ValueError(
                f"{axis=} must be an element of {values.shape}"
            )

    shape_values = na.shape(values)
    shape_edges = na.shape(edges)
    shape_args = na.broadcast_shapes(
        shape_values,
        {a: shape_edges[a] for a in shape_edges if a != axis},
    )

    shape = na.broadcast_shapes(na.shape(ax), shape_args)

    values = na.broadcast_to(values, shape)

    if edges is not None:
        edges = na.broadcast_to(
            array=edges,
            shape={a: shape[a] + 1 if a == axis else shape[a] for a in shape},
        )

    axis_distribution = values.axis_distribution
    dshape_edges = na.shape(edges.distribution) if edges is not None else dict()
    shape_distribution = na.broadcast_shapes(
        na.shape(values.distribution),
        {a: dshape_edges[a] for a in dshape_edges if a != axis},
    )

    if axis_distribution in shape_distribution:
        num_distribution = shape_distribution[axis_distribution]
    else:
        num_distribution = 1

    if num_distribution == 0:
        alpha = 1
    else:
        alpha = max(1 / num_distribution, 1/255)
    if "alpha" in kwargs:
        kwargs["alpha"] *= alpha
    else:
        kwargs["alpha"] = na.UncertainScalarArray(1, alpha)

    if "color" not in kwargs:
        color_cycle = plt.rcParams['axes.prop_cycle'].by_key()['color']
        shape_orthogonal = {a: shape[a] for a in shape if a != axis}
        color = na.ScalarArray.empty(shape=shape_orthogonal, dtype=object)
        for i, index in enumerate(color.ndindex()):
            color[index] = color_cycle[i % len(color_cycle)]
        kwargs["color"] = uncertainties._normalize(color)

    result = na.UncertainScalarArray(
        nominal=na.plt.stairs(
            edges.nominal if edges is not None else edges,
            values.nominal,
            ax=ax,
            axis=axis,
            where=where.nominal,
            **{k: kwargs[k].nominal for k in kwargs}
        ),
        distribution=na.plt.stairs(
            edges.distribution if edges is not None else edges,
            values.distribution,
            ax=ax,
            axis=axis,
            where=where.distribution,
            **{k: kwargs[k].distribution for k in kwargs}
        )
    )

    return result


@_implements(na.jacobian)
def jacobian(
        function: Callable[[na.AbstractScalar], na.AbstractScalar],
        x: na.AbstractScalar,
        dx: None | na.AbstractScalar = None,
        like: None | na.AbstractScalar = None,
) -> na.AbstractScalar:
    return named_arrays._scalars.scalar_named_array_functions.jacobian(
        function=function,
        x=x,
        dx=dx,
        like=like,
    )


@_implements(na.optimize.root_newton)
def optimize_root_newton(
        function: Callable[[na.ScalarLike], na.ScalarLike],
        guess: na.ScalarLike,
        jacobian: Callable[[na.ScalarLike], na.ScalarLike],
        max_abs_error: na.ScalarLike,
        max_iterations: int = 100,
        callback: None | Callable[[int, na.ScalarLike, na.ScalarLike, na.ScalarLike], None] = None,
) -> na.UncertainScalarArray:
    return named_arrays._scalars.scalar_named_array_functions.optimize_root_newton(
        function=function,
        guess=guess,
        jacobian=jacobian,
        max_abs_error=max_abs_error,
        max_iterations=max_iterations,
        callback=callback,
    )


@_implements(na.optimize.root_secant)
def optimize_root_secant(
        function: Callable[[na.ScalarLike], na.ScalarLike],
        guess: na.ScalarLike,
        min_step_size: na.ScalarLike,
        max_abs_error: na.ScalarLike,
        max_iterations: int = 100,
        damping: None | float = None,
        callback: None | Callable[[int, na.ScalarLike, na.ScalarLike, na.ScalarLike], None] = None,
) -> na.UncertainScalarArray:

    try:
        guess = uncertainties._normalize(guess).astype(float)

        min_step_size = uncertainties._normalize(min_step_size)

        x0 = guess - 10 * min_step_size
        x1 = guess

        f0 = function(x0)
        f0 = uncertainties._normalize(f0)

        max_abs_error = uncertainties._normalize(max_abs_error)

    except uncertainties.UncertainScalarTypeError:
        return NotImplemented

    if na.shape(max_abs_error):
        raise ValueError(f"argument `max_abs_error` should have an empty shape, got {na.shape(max_abs_error)}")

    shape = na.shape_broadcasted(f0, guess, min_step_size)
    shape_distribution = na.broadcast_shapes(
        f0.shape_distribution,
        guess.shape_distribution,
        min_step_size.shape_distribution,
    )

    f0 = na.broadcast_to(f0, shape)
    f0.distribution = na.broadcast_to(f0.distribution, shape_distribution)

    converged = na.broadcast_to(0 * na.value(f0), shape=shape).astype(bool)

    x1 = na.broadcast_to(x1, shape)
    x1.distribution = na.broadcast_to(x1.distribution, shape_distribution)

    for i in range(max_iterations):

        f1 = function(x1)

        if callback is not None:
            callback(i, x1, f1, converged)

        if max_abs_error is not None:
            converged |= np.abs(f1) < max_abs_error

        dx = x1 - x0

        converged |= np.abs(dx) < np.abs(min_step_size)

        if np.all(converged):
            return x1

        active = ~converged

        dx_active = na.UncertainScalarArray(
            nominal=dx.nominal[active.nominal],
            distribution=dx.distribution[active.distribution],
        )
        f0_active = na.UncertainScalarArray(
            nominal=f0.nominal[active.nominal],
            distribution=f0.distribution[active.distribution],
        )
        f1_active = na.UncertainScalarArray(
            nominal=f1.nominal[active.nominal],
            distribution=f1.distribution[active.distribution],
        )

        df_active = f1_active - f0_active
        if np.any(df_active.nominal == 0) or np.any(df_active.distribution == 0):
            raise ValueError("stationary point detected")

        jacobian = df_active / dx_active

        correction = f1_active / jacobian
        if damping is not None:
            correction = damping * correction

        x2 = x1.copy()
        x2.nominal[active.nominal] = x2.nominal[active.nominal] - correction.nominal
        x2.distribution[active.distribution] = x2.distribution[active.distribution] - correction.distribution

        x0 = x1
        x1 = x2
        f0 = f1

    raise ValueError("Max iterations exceeded")


def _colorsynth_normalize(
    spd: na.AbstractScalar,
    wavelength: None | na.AbstractScalar = None,
    axis: None | str = None,
    spd_min: None | na.ScalarLike = None,
    spd_max: None | na.ScalarLike = None,
    wavelength_min: None | na.ScalarLike = None,
    wavelength_max: None | na.ScalarLike = None,
) -> tuple[
    na.AbstractUncertainScalarArray,
    na.AbstractUncertainScalarArray,
    str,
    na.AbstractUncertainScalarArray,
    na.AbstractUncertainScalarArray,
    na.AbstractUncertainScalarArray,
    na.AbstractUncertainScalarArray,
]:

    try:
        spd = uncertainties._normalize(spd)
    except na.UncertainScalarArray:     # pragma: nocover
        return NotImplemented

    spd = spd.broadcasted
    shape = spd.shape

    if axis is None:
        axes = tuple(set(na.shape(spd)) | set(na.shape(wavelength)))
        if len(axes) != 1:
            raise ValueError(
                f"If `axis` is `None`, the other arguments must have zero "
                f"or one axis, got {axes}."
            )
        else:
            axis = axes[0]

    if wavelength is None:
        wavelength = na.linspace(0, 1, axis=axis, num=shape[axis])

    try:
        wavelength = uncertainties._normalize(wavelength)
    except na.UncertainScalarArray:     # pragma: nocover
        return NotImplemented

    if spd_min is None:
        spd_min = np.minimum(
            spd.nominal.min(),
            spd.distribution.min(),
        )
    if spd_max is None:
        spd_max = np.maximum(
            spd.nominal.max(),
            spd.distribution.max(),
        )
    if wavelength_min is None:
        wavelength_min = np.minimum(
            wavelength.nominal.min(),
            wavelength.distribution.min(),
        )
    if wavelength_max is None:
        wavelength_max = np.maximum(
            wavelength.nominal.max(),
            wavelength.distribution.max(),
        )

    try:
        spd_min = uncertainties._normalize(spd_min)
        spd_max = uncertainties._normalize(spd_max)
        wavelength_min = uncertainties._normalize(wavelength_min)
        wavelength_max = uncertainties._normalize(wavelength_max)
    except na.UncertainScalarArray:     # pragma: nocover
        return NotImplemented

    return (
        spd,
        wavelength,
        axis,
        spd_min,
        spd_max,
        wavelength_min,
        wavelength_max,
    )

@_implements(na.colorsynth.rgb)
def colorsynth_rgb(
    spd: na.AbstractScalar,
    wavelength: None | na.AbstractScalar = None,
    axis: None | str = None,
    spd_min: None | na.ScalarLike = None,
    spd_max: None | na.ScalarLike = None,
    spd_norm: None | Callable = None,
    wavelength_min: None | na.ScalarLike = None,
    wavelength_max: None | na.ScalarLike = None,
    wavelength_norm: None | Callable = None,
) -> na.UncertainScalarArray:

    (
        spd,
        wavelength,
        axis,
        spd_min,
        spd_max,
        wavelength_min,
        wavelength_max,
    ) = _colorsynth_normalize(
        spd,
        wavelength,
        axis,
        spd_min,
        spd_max,
        wavelength_min,
        wavelength_max,
    )

    return na.UncertainScalarArray(
        nominal=na.colorsynth.rgb(
            spd=spd.nominal,
            wavelength=wavelength.nominal,
            axis=axis,
            spd_min=spd_min.nominal,
            spd_max=spd_max.nominal,
            spd_norm=spd_norm,
            wavelength_min=wavelength_min.nominal,
            wavelength_max=wavelength_max.nominal,
            wavelength_norm=wavelength_norm,
        ),
        distribution=na.colorsynth.rgb(
            spd=spd.distribution,
            wavelength=wavelength.distribution,
            axis=axis,
            spd_min=spd_min.distribution,
            spd_max=spd_max.distribution,
            spd_norm=spd_norm,
            wavelength_min=wavelength_min.distribution,
            wavelength_max=wavelength_max.distribution,
            wavelength_norm=wavelength_norm,
        ),
    )


@_implements(na.colorsynth.colorbar)
def colorsynth_colorbar(
    spd: na.AbstractScalar,
    wavelength: None | na.AbstractScalar = None,
    axis: None | str = None,
    spd_min: None | na.ScalarLike = None,
    spd_max: None | na.ScalarLike = None,
    spd_norm: None | Callable = None,
    wavelength_min: None | na.ScalarLike = None,
    wavelength_max: None | na.ScalarLike = None,
    wavelength_norm: None | Callable = None,
) -> na.FunctionArray[na.Cartesian2dVectorArray, na.UncertainScalarArray]:

    (
        spd,
        wavelength,
        axis,
        spd_min,
        spd_max,
        wavelength_min,
        wavelength_max,
    ) = _colorsynth_normalize(
        spd,
        wavelength,
        axis,
        spd_min,
        spd_max,
        wavelength_min,
        wavelength_max,
    )

    result_nominal = na.colorsynth.colorbar(
        spd=na.as_named_array(spd.nominal),
        wavelength=wavelength.nominal,
        axis=axis,
        spd_min=spd_min.nominal,
        spd_max=spd_max.nominal,
        spd_norm=spd_norm,
        wavelength_min=wavelength_min.nominal,
        wavelength_max=wavelength_max.nominal,
        wavelength_norm=wavelength_norm,
    )

    result_distribution = na.colorsynth.colorbar(
        spd=na.as_named_array(spd.distribution),
        wavelength=wavelength.distribution,
        axis=axis,
        spd_min=spd_min.distribution,
        spd_max=spd_max.distribution,
        spd_norm=spd_norm,
        wavelength_min=wavelength_min.distribution,
        wavelength_max=wavelength_max.distribution,
        wavelength_norm=wavelength_norm,
    )

    return na.FunctionArray(
        inputs=na.Cartesian2dVectorArray(
            x=na.UncertainScalarArray(
                nominal=result_nominal.inputs.x,
                distribution=result_distribution.inputs.x,
            ),
            y=na.UncertainScalarArray(
                nominal=result_nominal.inputs.y,
                distribution=result_distribution.inputs.y,
            ),
        ),
        outputs=na.UncertainScalarArray(
            nominal=result_nominal.outputs,
            distribution=result_distribution.outputs,
        )
    )


def ndfilter(
    func: Callable,
    array: na.AbstractScalar,
    size: dict[str, int],
    where: na.AbstractScalar,
    **kwargs,
) -> na.UncertainScalarArray:

    try:
        array = uncertainties._normalize(array)
        where = uncertainties._normalize(where)
    except uncertainties.UncertainScalarTypeError:  # pragma: nocover
        return NotImplemented

    array = array.broadcasted
    where = where.broadcasted

    return array.type_explicit(
        nominal=func(
            array=array.nominal,
            size=size,
            where=where.nominal,
            **kwargs,
        ),
        distribution=func(
            array=array.distribution,
            size=size,
            where=where.distribution,
            **kwargs,
        )
    )


@_implements(na.despike)
def despike(
    array: na.AbstractScalar,
    axis: tuple[str, str],
    where: None | bool | na.AbstractScalar,
    inbkg: None | na.AbstractScalar,
    invar: None | float | na.AbstractScalar,
    sigclip: float,
    sigfrac: float,
    objlim: float,
    gain: float,
    readnoise: float,
    satlevel: float,
    niter: int,
    sepmed: bool,
    cleantype: Literal["median", "medmask", "meanmask", "idw"],
    fsmode: Literal["median", "convolve"],
    psfmodel: Literal["gauss", "gaussx", "gaussy", "moffat"],
    psffwhm: float,
    psfsize: int,
    psfk: None | na.AbstractScalar,
    psfbeta: float,
    verbose: bool,
) -> na.ScalarArray:

    try:
        array = uncertainties._normalize(array)
        where = uncertainties._normalize(where)
        inbkg = uncertainties._normalize(inbkg)
        invar = uncertainties._normalize(invar)
        psfk = uncertainties._normalize(psfk)
    except uncertainties.UncertainScalarTypeError:  # pragma: nocover
        return NotImplemented

    kwargs = dict(
        axis=axis,
        sigclip=sigclip,
        sigfrac=sigfrac,
        objlim=objlim,
        gain=gain,
        readnoise=readnoise,
        satlevel=satlevel,
        niter=niter,
        sepmed=sepmed,
        cleantype=cleantype,
        fsmode=fsmode,
        psfmodel=psfmodel,
        psffwhm=psffwhm,
        psfsize=psfsize,
        psfbeta=psfbeta,
        verbose=verbose,
    )

    result = array.copy_shallow()
    result.nominal = na.despike(
        array=array.nominal,
        where=where.nominal,
        inbkg=inbkg.nominal,
        invar=invar.nominal,
        psfk=psfk.nominal,
        **kwargs,
    )
    result.distribution = na.despike(
        array=array.distribution,
        where=where.distribution,
        inbkg=inbkg.distribution,
        invar=invar.distribution,
        psfk=psfk.distribution,
        **kwargs,
    )

    return result



from __future__ import annotations
from typing import Sequence, overload, Type, Any, Callable, TypeVar, Literal
import functools
import numpy as np
import astropy.units as u
import named_arrays as na

__all__ = [
    '_named_array_function',
    'asarray',
    'asanyarray',
    'arange',
    'step',
    'linspace',
    'logspace',
    'geomspace',
    'ndim',
    'shape',
    'unit',
    'unit_normalized',
    'broadcast_to',
    'stack',
    'concatenate',
    'add_axes',
    "vmr",
    "interp",
    "histogram",
    "histogram2d",
    "histogramdd",
    'jacobian',
    'despike',
]

ArrayT = TypeVar("ArrayT")
LikeT = TypeVar("LikeT", bound="None | na.AbstractArray")
AxisT = TypeVar("AxisT", bound="str | na.AbstractArray")
NumT = TypeVar("NumT", bound="int | na.AbstractArray")
BaseT = TypeVar("BaseT", bound="int | na.AbstractArray")
InputT = TypeVar("InputT", bound="float | u.Quantity | na.AbstractScalarArray")
OutputT = TypeVar("OutputT", bound="float | u.Quantity | na.AbstractScalarArray")
WhereT = TypeVar("WhereT", bound="bool | na.AbstractScalarArray")


def _is_subclass(a: Any, b: Any):
    if type(a) == type(b):
        return 0
    elif isinstance(a, type(b)):
        return 1
    elif isinstance(b, type(a)):
        return -1
    else:
        return 0


def _named_array_function(func: Callable, *args, **kwargs):
    arrays_args = tuple(arg for arg in args if isinstance(arg, na.AbstractArray))
    arrays_kwargs = tuple(kwargs[k] for k in kwargs if isinstance(kwargs[k], na.AbstractArray))
    arrays = arrays_args + arrays_kwargs

    arrays = sorted(arrays, key=functools.cmp_to_key(_is_subclass))

    for array in arrays:
        res = array.__named_array_function__(func, *args, **kwargs)
        if res is not NotImplemented:
            return res

    raise TypeError(
        f"all types, {[type(a).__qualname__ for a in arrays]}, returned `NotImplemented` "
        f"for function `{func.__name__}`"
    )


def _asarray_like(
        func: Callable,
        a: ArrayT,
        dtype: None | type | np.dtype | str = None,
        order: None | str = None,
        *,
        like: None | LikeT = None,
) -> ArrayT | LikeT:
    if a is None:
        return None

    if like is None:
        like = na.ScalarArray(None)
    elif not na.named_array_like(like):
        like = func(like)

    return _named_array_function(
        func=func,
        a=a,
        dtype=dtype,
        order=order,
        like=like,
    )


@overload
def asarray(
        a: ArrayT,
        dtype: None | type | np.dtype | str = ...,
        order: None | str = ...,
        *,
        like: None = ...
) -> ArrayT:
    ...


@overload
def asarray(
        a: ArrayT,
        dtype: None | type | np.dtype | str = ...,
        order: None | str = ...,
        *,
        like: LikeT = ...
) -> LikeT:
    ...


def asarray(
        a: ArrayT,
        dtype: None | type | np.dtype | str = None,
        order: None | str = None,
        *,
        like: None | LikeT = None,
) -> ArrayT | LikeT:
    """
    Converts the input to use only instances of :class:`numpy.ndarray` as the underlying data.

    This function does not convert an instance of :class:`named_arrays.AbstractArray` to an instance of
    :class:`numpy.ndarray` like you might expect from the documentation of :func:`numpy.asarray`.
    Instead, it recursively inspects the input, converting instances of :class:`named_arrays.AbstractImplicitArray`
    to `named_arrays.AbstractExplicitArray`, and calling :func:`numpy.asarray` on the underlying data.

    Parameters
    ----------
    a
        Input array to be converted
    dtype
        Data type of output, usually inferred from the input.
    order
        Memory layout. See the documentation of :func:`numpy.asarray` for more information.
    like
        Optional reference object.
        If provided, the result will be defined by this object.

    Returns
    -------
    out
        Standardized interpretation of ``a``, with all the underlying data expressed as instances of
        :class:`numpy.ndarray`.

    Examples
    --------

    Standardize a :class:`float`

    .. jupyter-execute::

        import named_arrays as na

        na.asarray(2)

    Standardize a :class:`named_arrays.ScalarArray` of :class:`float`

    .. jupyter-execute::

        na.asarray(na.ScalarArray(2))

    Standardize a :class:`named_arrays.Cartesian2dVectorArray` of :class:`float`

    .. jupyter-execute::

        na.asarray(na.Cartesian2dVectorArray(2, 3))

    See Also
    --------
    :func:`numpy.asarray` : Equivalent Numpy function
    :func:`named_arrays.asanyarray` : Similar to this function, but allows `numpy.ndarray` subclasses to pass through.
    """
    return _asarray_like(
        func=asarray,
        a=a,
        dtype=dtype,
        order=order,
        like=like,
    )


@overload
def asanyarray(
        a: ArrayT,
        dtype: None | type | np.dtype | str = ...,
        order: None | str = ...,
        *,
        like: None = ...
) -> ArrayT:
    ...


@overload
def asanyarray(
        a: ArrayT,
        dtype: None | type | np.dtype | str = ...,
        order: None | str = ...,
        *,
        like: LikeT = ...
) -> LikeT:
    ...


def asanyarray(
        a: ArrayT,
        dtype: None | type | np.dtype | str = None,
        order: None | str = None,
        *,
        like: None | LikeT = None,
) -> ArrayT | LikeT:
    """
    Converts the input to use only instances of :class:`numpy.ndarray` subclasses as the underlying data.

    This function does not convert an instance of :class:`named_arrays.AbstractArray` to an instance of a
    :class:`numpy.ndarray` subclass like you might expect from the documentation of :func:`numpy.asanyarray`.
    Instead, it recursively inspects the input, converting instances of :class:`named_arrays.AbstractImplicitArray`
    to `named_arrays.AbstractExplicitArray`, and calling :func:`numpy.asanyarray` on the underlying data.

    Parameters
    ----------
    a
        Input array to be converted
    dtype
        Data type of output, usually inferred from the input.
    order
        Memory layout. See the documentation of :func:`numpy.asanyarray` for more information.
    like
        Optional reference object.
        If provided, the result will be defined by this object.

    Returns
    -------
    out
        Standardized interpretation of ``a``, with all the underlying data expressed as instances of
        :class:`numpy.ndarray` subclasses.

    Examples
    --------

    Standardize an instance of :class:`astropy.units.Quantity`

    .. jupyter-execute::

        import astropy.units as u
        import named_arrays as na

        na.asanyarray(2 * u.mm)

    See Also
    --------
    :func:`numpy.asanyarray` : Equivalent Numpy function
    :func:`named_arrays.asarray`: Similar to this function but converts instances of :class:`numpy.ndarray` subclasses
        back to instances of :class:`numpy.ndarray`.
    """

    return _asarray_like(
        func=asanyarray,
        a=a,
        dtype=dtype,
        order=order,
        like=like,
    )


def arange(
        start: float | complex | u.Quantity | na.AbstractArray,
        stop: float | complex | u.Quantity | na.AbstractArray,
        axis: str | na.AbstractArray,
        step: int | na.AbstractArray = 1,
) -> na.AbstractExplicitArray:
    """
    Redefined version of :func:`numpy.arange` with an ``axis`` parameter.

    Parameters
    ----------
    start
        starting value of the sequence
    stop
        ending value of the sequence
    axis
        name of the new sequence axis
    step
        step size between consecutive elements of the sequence

    See Also
    --------
    :func:`numpy.arange` : Equivalent numpy function

    :class:`named_arrays.ScalarArrayRange` : Implicit array version of this function for scalars
    """
    return _named_array_function(
        func=arange,
        start=na.as_named_array(start),
        stop=na.as_named_array(stop),
        axis=axis,
        step=step,
    )


def step(
    start: na.StartT,
    stop: na.StopT,
    num: NumT,
    endpoint: bool = True,
    centers: bool = False,
) -> na.StartT | na.StopT | NumT:
    """
    Helper function to compute the step size for :func:`linspace`.

    Parameters
    ----------
    start
        The starting value of the sequence.
    stop
        The last value of the sequence, unless :attr:`endpoint` is :class:`False`.
    num
        The number of values in the sequence.
    endpoint
        Flag controlling whether :attr:`stop` should be included in the sequence.
    centers
        Flag controlling whether the returned values should be on cell centers.
        The default is to return values on cell edges.
    """
    if endpoint:
        num = num - 1
    if centers:
        num = num + 1
    result = (stop - start) / num
    return result


def linspace(
        start: na.StartT,
        stop: na.StopT,
        axis: AxisT,
        num: NumT = 50,
        endpoint: bool = True,
        dtype: None | type | np.dtype = None,
        centers: bool = False,
) -> na.StartT | na.StopT | AxisT | NumT:
    """
    Create an array of evenly-spaced numbers between :attr:`start` and :attr:`stop`

    Parameters
    ----------
    start
        The starting value of the sequence
    stop
        The last value of the sequence, unless :attr:`endpoint` is :class:`False`
    axis
        The name of the new axis corresponding to the sequence
    num
        Number of values in the sequence
    endpoint
        Flag controlling whether :attr:`stop` should be included in the sequence.
    dtype
        :class:`numpy.dtype` of the result
    centers
        Flag controlling whether the returned values should be on cell centers.
        The default is to return values on cell edges.

    See Also
    --------
    :func:`numpy.linspace` :  Corresponding numpy function.

    :class:`named_arrays.ScalarLinearSpace` : Corresponding implicit scalar array.

    :class:`named_arrays.UncertainScalarLinearSpace`: Corresponding implicit uncertain scalar array.

    Examples
    --------

    Compare the points returned using ``centers=False`` to those returned
    using ``centers=True``.

    .. jupyter-execute::

        import named_arrays as na
        import matplotlib.pyplot as plt

        # Define an array of cell edges
        edges_x = na.linspace(0, 1, axis="x", num=11)
        edges_y = na.linspace(0, 1, axis="y", num=11)

        # Define an array of cell centers
        centers_x = na.linspace(0, 1, axis="x", num=10, centers=True)
        centers_y = na.linspace(0, 1, axis="y", num=10, centers=True)

        # Plot the results as a scatterplot
        fig, ax = plt.subplots()
        na.plt.scatter(edges_x, edges_y, ax=ax, label="edges");
        na.plt.scatter(centers_x, centers_y, ax=ax, label="centers");
        ax.legend();
    """

    if centers:
        halfstep = step(
            start=start,
            stop=stop,
            num=num,
            endpoint=endpoint,
            centers=centers,
        ) / 2
        start = start + halfstep
        stop = stop - halfstep

    return np.linspace(
        start=na.as_named_array(start),
        stop=stop,
        axis=axis,
        num=num,
        endpoint=endpoint,
        dtype=dtype,
    )


def logspace(
        start: na.StartT,
        stop: na.StopT,
        axis: AxisT,
        num: NumT = 50,
        endpoint: bool = True,
        base: BaseT = 10,
        dtype: None | type | np.dtype = None,
) -> na.StartT | na.StopT | AxisT | NumT:
    """
    Create an array of evenly-spaced numbers on a log scale between :attr:`start` and :attr:`stop`

    Parameters
    ----------
    start
        The starting value of the sequence
    stop
        The last value of the sequence, unless :attr:`endpoint` is :class:`False`
    axis
        The name of the new axis corresponding to the sequence
    num
        Number of values in the sequence
    endpoint
        Flag controlling whether :attr:`stop` should be included in the sequence.
    base
        The base of the logarithm
    dtype
        :class:`numpy.dtype` of the result

    See Also
    --------
    :func:`numpy.logspace` :  Corresponding numpy function.

    :class:`named_arrays.ScalarLogarithmicSpace` : Corresponding implicit scalar array.

    :class:`named_arrays.UncertainScalarLogarithmicSpace`: Corresponding implicit uncertain scalar array.
    """
    return np.logspace(
        start=na.as_named_array(start),
        stop=stop,
        axis=axis,
        num=num,
        endpoint=endpoint,
        base=base,
        dtype=dtype,
    )


def geomspace(
        start: na.StartT,
        stop: na.StopT,
        axis: AxisT,
        num: NumT = 50,
        endpoint: bool = True,
        dtype: None | type | np.dtype = None,
) -> na.StartT | na.StopT | AxisT | NumT:
    """
    Create an array of a geometric progression of numbers between :attr:`start` and :attr:`stop`

    Parameters
    ----------
    start
        The starting value of the sequence
    stop
        The last value of the sequence, unless :attr:`endpoint` is :class:`False`
    axis
        The name of the new axis corresponding to the sequence
    num
        Number of values in the sequence
    endpoint
        Flag controlling whether :attr:`stop` should be included in the sequence.
    dtype
        :class:`numpy.dtype` of the result

    See Also
    --------
    :func:`numpy.geomspace` :  Corresponding numpy function.

    :class:`named_arrays.ScalarGeometricSpace` : Corresponding implicit scalar array.

    :class:`named_arrays.UncertainScalarGeometricSpace`: Corresponding implicit uncertain scalar array.
    """
    return np.geomspace(
        start=na.as_named_array(start),
        stop=stop,
        axis=axis,
        num=num,
        endpoint=endpoint,
        dtype=dtype,
    )


def ndim(a: na.AbstractArray) -> int:
    return np.ndim(a)


def shape(a: na.ArrayLike) -> dict[str, int]:
    if not isinstance(a, na.AbstractArray):
        a = na.ScalarArray(a)
    return np.shape(a)


def unit(
        a: Any,
        unit_dimensionless: None | float | u.UnitBase = None,
        squeeze: bool = True,
) -> None | u.UnitBase | na.AbstractArray:
    """
    Isolate the physical units associated with the given object.

    If the array has no physical units, this function returns :obj:`None`.

    Parameters
    ----------
    a
        object to isolate the units of
    unit_dimensionless
        The unit to use for dimensionless objects.
    squeeze
        If the result is an instance of :class:`named_arrays.AbstractVectorArray`,
        and all the components are the same, simplify the result into a single
        :class:`astropy.units.Unit` instance.

    See Also
    --------
    :func:`unit_normalized` : version of this function that by default returns
        :obj:`astropy.units.dimensionless_unscaled` instead of :obj:`None`
        if there is no unit associated with the given object.
    """
    if isinstance(a, u.UnitBase):
        return a
    elif isinstance(a, u.Quantity):
        return a.unit
    elif isinstance(a, na.AbstractArray):
        return na._named_array_function(
            func=unit,
            a=a,
            unit_dimensionless=unit_dimensionless,
            squeeze=squeeze,
        )
    else:
        return unit_dimensionless


def unit_normalized(
        a: Any,
        unit_dimensionless: float | u.UnitBase = u.dimensionless_unscaled,
        squeeze: bool = True,
) -> u.UnitBase | na.AbstractArray:
    """
    Isolate the physical units associated with a given object,`
    normalizing to the given dimensionless units if the object does not have
    associated units.


    Parameters
    ----------
    a
        object to isolate the units of
    unit_dimensionless
        The unit to use for dimensionless objects.
    squeeze
        If the result is an instance of :class:`named_arrays.AbstractVectorArray`,
        and all the components are the same, simplify the result into a single
        :class:`astropy.units.Unit` instance.

    See Also
    --------
    :func:`unit` : version of this function that by default returns :obj:`None`
        instead of :obj:`astropy.units.dimensionless_unscaled` if there is no
        unit associated with the given object.

    """
    if isinstance(a, u.UnitBase):
        return a
    elif isinstance(a, u.Quantity):
        return a.unit
    elif isinstance(a, na.AbstractArray):
        return na._named_array_function(
            func=unit_normalized,
            a=a,
            unit_dimensionless=unit_dimensionless,
            squeeze=squeeze,
        )
    else:
        return unit_dimensionless


@overload
def broadcast_to(array: float | complex | np.ndarray | u.Quantity, shape: dict[str, int]) -> na.ScalarArray:
    ...


@overload
def broadcast_to(array: na.AbstractScalarArray, shape: dict[str, int]) -> na.ScalarArray:
    ...


def broadcast_to(array, shape):
    if not isinstance(array, na.AbstractArray):
        array = na.ScalarArray(array)
    return np.broadcast_to(array=array, shape=shape)


def stack(
        arrays: Sequence[na.ArrayLike],
        axis: str,
        out: None | na.AbstractExplicitArray = None,
        *,
        dtype: str | np.dtype | Type = None,
        casting: None | str = "same_kind",
) -> na.AbstractArray:
    if not any(isinstance(a, na.AbstractArray) for a in arrays):
        arrays = list(arrays)
        arrays[0] = na.ScalarArray(arrays[0])
    return np.stack(
        arrays=arrays,
        axis=axis,
        out=out,
        dtype=dtype,
        casting=casting,
    )


def concatenate(
        arrays: Sequence[na.ArrayLike],
        axis: str,
        out: None | na.AbstractExplicitArray = None,
        *,
        dtype: str | np.dtype | Type = None,
        casting: None | str = "same_kind",
) -> na.AbstractArray:
    if not any(isinstance(a, na.AbstractArray) for a in arrays):
        arrays = list(arrays)
        arrays[0] = na.ScalarArray(arrays[0])
    return np.concatenate(
        arrays=arrays,
        axis=axis,
        out=out,
        dtype=dtype,
        casting=casting,
    )


def add_axes(array: na.ArrayLike, axes: str | Sequence[str]):
    if not isinstance(array, na.AbstractArray):
        array = na.ScalarArray(array)
    return array.add_axes(axes)


def vmr(
    a: ArrayT,
    axis: None | str | Sequence[str] = None,
    dtype: None | str | Type | np.dtype = None,
    out: None | na.AbstractExplicitArray = None,
    keepdims: bool = False,
    *,
    where: bool | WhereT = True,
) -> ArrayT | WhereT:
    """
    Compute the
    `variance-to-mean ratio <https://en.wikipedia.org/wiki/Index_of_dispersion>`_
    (also known as the `Fano factor <https://en.wikipedia.org/wiki/Fano_factor>`_)
    of the given array.

    Parameters
    ----------
    a
        Input array
    axis
        The axis or axes along which to compute the VMR.
        If :obj:`None` (the default), the VMR is computed along all the axes
        of the array.
    dtype
        The data type of the output
    out
        An optional output array in which to store the results.
    keepdims
        If :obj:`True`, the resulting array will have the same dimensionality.
    where
        A boolean mask indicating which elements to consider when computing the
        VMR.
    """

    kwargs = dict(
        a=a,
        axis=axis,
        dtype=dtype,
        keepdims=keepdims,
        where=where,
    )

    result = np.var(out=out, **kwargs)
    result = np.divide(result, np.mean(**kwargs), out=out)

    return result


def interp(
        x: float | u.Quantity | na.AbstractArray,
        xp: na.AbstractArray,
        fp: na.AbstractArray,
        axis: None | str = None,
        left: None | float | u.Quantity | na.AbstractArray = None,
        right: None | float | u.Quantity | na.AbstractArray = None,
        period: None | float | u.Quantity | na.AbstractArray = None,
) -> na.AbstractArray:
    """
    Thin wrapper around :func:`numpy.interp`.

    Performs 1D interpolation on monotonically-increasing sample points along
    the specified axes.

    This function adds an ``axis`` argument to allow for interpolating
    :math:`n`-dimensional arrays.

    Parameters
    ----------
    x
         The new :math:`x` coordinates where the interpolant will be evaluated.
    xp
        The :math:`x` coordinates of the data points.
    fp
        The :math:`y` coordinates of the data points.
    axis
        The logical axis along which to interpolate.
    left
        Value to return for points less than ``xp[{axis: 0}]``.
        Default is ``fp[{axis: 0}]``
    right
        Value to return for points larger than ``xp[{axis: ~0}]``
    period
        A period for the :math:`x` coordinates.
        This parameter allows for proper interpolation of angular coordinates
    """
    return _named_array_function(
        func=interp,
        x=x,
        xp=xp,
        fp=fp,
        axis=axis,
        left=left,
        right=right,
        period=period,
    )


def histogram(
    a: na.AbstractArray,
    bins: dict[str, int] | na.AbstractArray,
    axis: None | str | Sequence[str] = None,
    min: None | float | na.AbstractArray = None,
    max: None | float | na.AbstractArray = None,
    density: bool = False,
    weights: None | na.AbstractArray = None,
) -> na.FunctionArray[na.AbstractArray, na.ScalarArray]:
    """
    A thin wrapper around :func:`numpy.histogram` which adds an `axis` argument.

    Parameters
    ----------
    a
        The input data over which to compute the histogram.
    bins
        The bin specification of the histogram:
         * If `bins` is a dictionary, the keys are interpreted as the axis names
           and the values are the number of bins along each axis.
           This dictionary must have only one key per coordinate.
         * If `bins` is an array, it represents the bin edges.
    axis
        The logical axes along which to histogram the data points.
        If :obj:`None` (the default), the histogram will be computed along
        all the axes of `a`.
    min
        The lower boundary of the histogram.
        If :obj:`None` (the default), the minimum of `a` is used.
    max
        The upper boundary of the histogram.
        If :obj:`None` (the default), the maximum of `a` is used.
    density
        If :obj:`False` (the default), returns the number of samples in each bin.
        If :obj:`True`, returns the probability density in each bin.
    weights
        An optional array weighting each sample.

    Examples
    --------

    Construct a 2D histogram with constant bin width.

    .. jupyter-execute::

        import numpy as np
        import matplotlib.pyplot as plt
        import named_arrays as na

        # Define the bin edges
        bins = dict(x=6)

        # Define random points to collect into a histogram
        a = na.random.normal(0, 2, shape_random=dict(h=101))

        # Compute the histogram
        hist = na.histogram(a, bins=bins)

        # Plot the resulting histogram
        fig, ax = plt.subplots()
        na.plt.stairs(hist.inputs, hist.outputs);
    """
    return _named_array_function(
        func=histogram,
        a=a,
        axis=axis,
        bins=bins,
        min=min,
        max=max,
        density=density,
        weights=weights,
    )


def histogram2d(
    x: na.AbstractScalarArray,
    y: na.AbstractScalarArray,
    bins: dict[str, int] | na.AbstractCartesian2dVectorArray,
    axis: None | str | Sequence[str] = None,
    min: None | na.AbstractScalarArray | na.AbstractCartesian2dVectorArray = None,
    max: None | na.AbstractScalarArray | na.AbstractCartesian2dVectorArray = None,
    density: bool = False,
    weights: None | na.AbstractScalarArray = None,
) -> na.FunctionArray[na.Cartesian2dVectorArray, na.ScalarArray]:
    """
    A thin wrapper around :func:`numpy.histogram2d` which adds an `axis` argument.

    Parameters
    ----------
    x
        An array containing the x coordinates of the points to be sampled.
    y
        An array containing the y coordinates of the points to be sampled.
    bins
        The bin specification of the histogram:
         * If `bins` is a dictionary, the keys are interpreted as the axis names
           and the values are the number of bins along each axis.
           This dictionary must have exactly two keys.
         * If `bins` is a 2D Cartesian vector, each component of the vector
           represents the bin edges in each dimension.
    axis
        The logical axes along which to histogram the data points.
        If :obj:`None` (the default), the histogram will be computed along
        all the axes of `x` and `y`.
    min
        The lower boundary of the histogram along each dimension.
        If :obj:`None` (the default), the minimum of `x` and `y` is used.
    max
        The upper boundary of the histogram along each dimension.
        If :obj:`None` (the default), the maximum of `x` and `y` is used.
    density
        If :obj:`False` (the default), returns the number of samples in each bin.
        If :obj:`True`, returns the probability density in each bin.
    weights
        An optional array weighting each sample.

    Examples
    --------

    Construct a 2D histogram with constant bin width.

    .. jupyter-execute::

        import numpy as np
        import matplotlib.pyplot as plt
        import named_arrays as na

        # Define the bin edges
        bins = dict(x=6, y=5)

        # Define random points to collect into a histogram
        x = na.random.normal(0, 2, shape_random=dict(h=101))
        y = na.random.normal(0, 3, shape_random=dict(h=101))

        # Compute the 2D histogram
        hist = na.histogram2d(x, y, bins=bins)

        # Plot the resulting histogram
        fig, ax = plt.subplots()
        na.plt.pcolormesh(C=hist);

    Construct a 2D histogram with variable bin width.

    .. jupyter-execute::

        import matplotlib.pyplot as plt
        import named_arrays as na

        # Define the bin edges
        bins = na.Cartesian2dVectorArray(
            x=np.square(na.linspace(0, 2, axis="x", num=6)),
            y=np.square(na.linspace(0, 2, axis="y", num=5)),
        )

        # Define random points to collect into a histogram
        x = na.random.normal(0, 2, shape_random=dict(h=101))
        y = na.random.normal(0, 3, shape_random=dict(h=101))

        # Compute the 2D histogram
        hist = na.histogram2d(x, y, bins=bins)

        # Plot the resulting histogram
        fig, ax = plt.subplots()
        na.plt.pcolormesh(C=hist);
    """
    return _named_array_function(
        func=histogram2d,
        x=x,
        y=y,
        axis=axis,
        bins=bins,
        min=min,
        max=max,
        density=density,
        weights=weights,
    )


def histogramdd(
    *sample: na.AbstractScalar,
    bins: dict[str, int] | na.AbstractScalar| Sequence[na.AbstractScalar],
    axis: None | str | Sequence[str] = None,
    min: None | na.AbstractScalar | Sequence[na.AbstractScalar] = None,
    max: None | na.AbstractScalar | Sequence[na.AbstractScalar] = None,
    density: bool = False,
    weights: None | na.AbstractScalar = None,
) -> tuple[na.AbstractScalar, tuple[na.AbstractScalar, ...]]:
    """
    A thin wrapper around :func:`numpy.histogramdd` which adds an `axis` argument.

    Parameters
    ----------
    sample
        The data to be histrogrammed.
        Note the difference in signature compared to :func:`numpy.histogramdd`,
        each component must be a separate argument,
        instead of a single argument containing a sequence of arrays.
        This is done so that multiple dispatch works better for this function.
    bins
        The bin specification of the histogram:
         * If `bins` is a dictionary, the keys are interpreted as the axis names
           and the values are the number of bins along each axis.
           This dictionary must have the same number of elements as `sample`.
         * If `bins` is an array or a sequence of arrays, it describes the
           monotonically-increasing bin edges along each dimension
    axis
        The logical axes along which to histogram the data points.
        If :obj:`None` (the default), the histogram will be computed along
        all the axes of `sample`.
    min
        The lower boundary of the histogram along each dimension.
        If :obj:`None` (the default), the minimum of each element of `sample` is used.
    max
        The upper boundary of the histogram along each dimension.
        If :obj:`None` (the default), the maximum of each elemennt of `sample` is used.
    density
        If :obj:`False` (the default), returns the number of samples in each bin.
        If :obj:`True`, returns the probability density in each bin.
    weights
        An optional array weighting each sample.
    """
    return _named_array_function(
        histogramdd,
        *[na.as_named_array(s) for s in sample],
        axis=axis,
        bins=bins,
        min=min,
        max=max,
        density=density,
        weights=weights,
    )


def jacobian(
        function: Callable[[InputT], OutputT],
        x: InputT,
        dx: None | InputT = None,
) -> na.AbstractMatrixArray:
    """
    Compute the Jacobian of the given function using the first-order finite difference method.

    Parameters
    ----------
    function
        The function to compute the Jacobian of
    x
        The point to evaluate the function
    dx
        The distance that ``x`` will be perturbed by to compute the slope
    """

    if dx is None:
        dx = 1e-10
        unit_x = na.unit(x)
        if unit_x is not None:
            dx = dx * unit_x
    dx = na.asanyarray(dx, like=x)

    f_x = function(x)

    return na._named_array_function(
        func=jacobian,
        function=function,
        x=x,
        dx=dx,
        like=f_x,
    )


def despike(
    array: ArrayT,
    axis: tuple[str, str],
    where: None | bool | na.AbstractArray = None,
    inbkg: None | na.AbstractArray = None,
    invar: None | float | ArrayT = None,
    sigclip: float = 4.5,
    sigfrac: float = 0.3,
    objlim: float = 5.0,
    gain: float = 1.0,
    readnoise: float = 6.5,
    satlevel: float = 65536.0,
    niter: int = 4,
    sepmed: bool = True,
    cleantype: Literal["median", "medmask", "meanmask", "idw"] = "meanmask",
    fsmode: Literal["median", "convolve"] = "median",
    psfmodel: Literal["gauss", "gaussx", "gaussy", "moffat"] = "gauss",
    psffwhm: float = 2.5,
    psfsize: int = 7,
    psfk: None | na.AbstractArray = None,
    psfbeta: float = 4.765,
    verbose: bool = False,
) -> ArrayT:
    """
    A thin wrapper around :func:`astroscrappy.detect_cosmics`
    :cite:t:`vanDokkum2001`, which removes cosmic ray spikes from a series of
    images.

    Parameters
    ----------
    array
        Input data array that will be used for cosmic ray detection. This
        should include the sky background (or a mean background level, added
        back in after sky subtraction), so that noise can be estimated
        correctly from the data values. This should be in units of "counts".
    axis
        The two axes defining the logical axes of each image.
    where
        A boolean array of which pixels to consider during the cleaning
        process. The inverse of `inmask` used in
        :func:`astroscrappy.detect_cosmics`.
    inbkg
        A pre-determined background image, to be subtracted from `array`
        before running the main detection algorithm.
        This is used primarily with spectroscopic data, to remove
        sky lines and the cross-section of an object continuum during
        iteration, "protecting" them from spurious rejection (see the above
        paper). This background is not removed from the final, cleaned output
        (`cleanarr`). This should be in units of "counts", the same units of `array`.
        This inbkg should be free from cosmic rays. When estimating the cosmic-ray
        free noise of the image, we will treat ``inbkg`` as a constant Poisson
        contribution to the variance.
    invar
        A pre-determined estimate of the data variance (ie. noise squared) in
        each pixel, generated by previous processing of `array`. If provided,
        this is used in place of an internal noise model based on `array`,
        ``gain`` and ``readnoise``. This still gets median filtered and cleaned
        internally, to estimate what the noise in each pixel *would* be in the
        absence of cosmic rays. This should be in units of "counts" squared.
    sigclip
        Laplacian-to-noise limit for cosmic ray detection. Lower values will
        flag more pixels as cosmic rays. Default: 4.5.
    sigfrac
        Fractional detection limit for neighboring pixels. For cosmic ray
        neighbor pixels, a lapacian-to-noise detection limit of
        sigfrac * sigclip will be used. Default: 0.3.
    objlim
        Minimum contrast between Laplacian image and the fine structure image.
        Increase this value if cores of bright stars are flagged as cosmic
        rays. Default: 5.0.
    gain
        Gain of the image (electrons / ADU). We always need to work in
        electrons for cosmic ray detection. Default: 1.0
    readnoise
        Read noise of the image (electrons). Used to generate the noise model
        of the image. Default: 6.5.
    satlevel
        Saturation of level of the image (electrons). This value is used to
        detect saturated stars and pixels at or above this level are added to
        the mask. Default: 65536.0.
    niter
        Number of iterations of the LA Cosmic algorithm to perform. Default: 4.
    sepmed
        Use the separable median filter instead of the full median filter.
        The separable median is not identical to the full median filter, but
        they are approximately the same and the separable median filter is
        significantly faster and still detects cosmic rays well. Default: True
    cleantype
        Set which clean algorithm is used:\n
        'median': An umasked 5x5 median filter\n
        'medmask': A masked 5x5 median filter\n
        'meanmask': A masked 5x5 mean filter\n
        'idw': A masked 5x5 inverse distance weighted interpolation\n
        Default: "meanmask".
    fsmode
        Method to build the fine structure image:\n
        'median': Use the median filter in the standard LA Cosmic algorithm
        'convolve': Convolve the image with the psf kernel to calculate the
        fine structure image using a matched filter technique.
        Default: 'median'.
    psfmodel
        Model to use to generate the psf kernel if fsmode == 'convolve' and
        psfk is None. The current choices are Gaussian and Moffat profiles.
        'gauss' and 'moffat' produce circular PSF kernels. The 'gaussx' and
        'gaussy' produce Gaussian kernels in the x and y directions
        respectively. Default: "gauss".
    psffwhm
        Full Width Half Maximum of the PSF to use to generate the kernel.
        Default: 2.5.
    psfsize
        Size of the kernel to calculate. Returned kernel will have size
        psfsize x psfsize. psfsize should be odd. Default: 7.
    psfk
        PSF kernel array to use for the fine structure image if
        fsmode == 'convolve'. If None and fsmode == 'convolve', we calculate
        the psf kernel using 'psfmodel'. Default: None.
    psfbeta
        Moffat beta parameter. Only used if fsmode=='convolve' and
        psfmodel=='moffat'. Default: 4.765.
    verbose
        Print to the screen or not. Default: False.
    """
    return na._named_array_function(
        despike,
        array=array,
        axis=axis,
        where=where,
        inbkg=inbkg,
        invar=invar,
        sigclip=sigclip,
        sigfrac=sigfrac,
        objlim=objlim,
        gain=gain,
        readnoise=readnoise,
        satlevel=satlevel,
        niter=niter,
        sepmed=sepmed,
        cleantype=cleantype,
        fsmode=fsmode,
        psfmodel=psfmodel,
        psffwhm=psffwhm,
        psfsize=psfsize,
        psfk=psfk,
        psfbeta=psfbeta,
        verbose=verbose,
    )



import pytest
import numpy as np
import astropy.units as u
import named_arrays as na


@pytest.mark.parametrize(
    argnames="n",
    argvalues=[
        10,
        (11 * u.photon).astype(int),
        na.ScalarArray(12),
        (na.arange(1, 10, axis="x") << u.photon).astype(int),
        na.Cartesian2dVectorArray(10, 11),
    ],
)
@pytest.mark.parametrize(
    argnames="p",
    argvalues=[
        0.5,
        na.ScalarArray(0.51),
        na.linspace(0.4, 0.5, axis="p", num=5),
        na.UniformUncertainScalarArray(0.5, width=0.1),
        na.Cartesian2dVectorArray(0.5, 0.6),
    ],
)
@pytest.mark.parametrize(
    argnames="shape_random",
    argvalues=[
        None,
        dict(_s=6),
    ],
)
@pytest.mark.parametrize(
    argnames="seed",
    argvalues=[
        None,
        42,
    ],
)
def test_binomial(
    n: int | u.Quantity | na.AbstractScalar | na.AbstractVectorArray,
    p: float | na.AbstractScalar | na.AbstractVectorArray,
    shape_random: None | dict[str, int],
    seed: None | int,
):
    result = na.random.binomial(
        n=n,
        p=p,
        shape_random=shape_random,
        seed=seed,
    )

    assert na.unit(result) == na.unit(n)

    assert np.all(result >= 0)
    assert np.all(result <= n)


@pytest.mark.parametrize(
    argnames="shape",
    argvalues=[
        0.5,
        na.ScalarArray(0.51),
        na.linspace(0.4, 0.5, axis="p", num=5),
        na.UniformUncertainScalarArray(0.5, width=0.1),
        na.Cartesian2dVectorArray(0.5, 0.6),
    ],
)
@pytest.mark.parametrize(
    argnames="scale",
    argvalues=[
        10,
        (11 * u.photon).astype(int),
        na.ScalarArray(12),
        (na.arange(1, 10, axis="x") << u.photon).astype(int),
        na.Cartesian2dVectorArray(10, 11),
    ],
)
@pytest.mark.parametrize(
    argnames="shape_random",
    argvalues=[
        None,
        dict(_s=6),
    ],
)
@pytest.mark.parametrize(
    argnames="seed",
    argvalues=[
        None,
        42,
    ],
)
def test_gamma(
    shape: float | na.AbstractScalar | na.AbstractVectorArray,
    scale: float | u.Quantity | na.AbstractScalar | na.AbstractVectorArray,
    shape_random: None | dict[str, int],
    seed: None | int,
):
    result = na.random.gamma(
        shape=shape,
        scale=scale,
        shape_random=shape_random,
        seed=seed,
    )

    assert na.unit(result) == na.unit(scale)
    assert np.all(result >= 0)



from __future__ import annotations
from typing import Type

import dataclasses
import named_arrays as na

__all__ = [
    'AbstractTemporalSpectralMatrixArray',
    'TemporalSpectralMatrixArray',
]


@dataclasses.dataclass(eq=False, repr=False)
class AbstractTemporalSpectralMatrixArray(
    na.AbstractTemporalSpectralVectorArray,
):

    @property
    def type_abstract(self) -> Type[AbstractTemporalSpectralMatrixArray]:
        return AbstractTemporalSpectralMatrixArray

    @property
    def type_explicit(self) -> Type[TemporalSpectralMatrixArray]:
        return TemporalSpectralMatrixArray

    @property
    def type_vector(self) -> Type[na.TemporalSpectralVectorArray]:
        return na.TemporalSpectralVectorArray

    @property
    def determinant(self) -> na.ScalarLike:     # pragma: nocover
        return NotImplementedError


@dataclasses.dataclass(eq=False, repr=False)
class TemporalSpectralMatrixArray(
    na.TemporalSpectralVectorArray,
    AbstractTemporalSpectralMatrixArray,
    na.AbstractExplicitMatrixArray,
):
    pass
